# Copyright (c) 2008 Drew Smathers.
# See LICENSE for details
from pyglet.gl import *

from miru.imiru import ICamera
from miru.imiru import IWorldRenderStage, IDebuggingRenderStage
from miru.imiru import IBlittableRenderStage
from miru.imiru import ITranslationHandler, IRotationHandler
from miru.utils import glvec
from miru import components
from miru.common import PositionalMixin, options

from euclid import Vector3
from UserList import UserList
from copy import copy
import math


from zope.interface import implementer, implements
from twisted.python.components import registerAdapter

lightno = lambda i : eval('GL_LIGHT%d' % i)

for LIGHT_CT_MAX in range(1000):
    try:
        lightno(LIGHT_CT_MAX)
    except Exception, e:
        #print 'Max lights count supported : ', LIGHT_CT_MAX
        break

def _constrain_lights(ct):
    assert ct <= LIGHT_CT_MAX,\
            'No more than %d lights can exist in a lightgroup' % LIGHT_CT_MAX
    if ct > 8:
        from warnings import warn
        warn("You're using more than 8 lights - this is not likely suppoted on other systems")

class Camera(PositionalMixin):
    """Abstraction representing a camera - the viewer's perspective.
    The also acts a C{IRenderStage} for world objects which are drawn
    after the initial tranformation matrix is applied.

    @param wireframe: draw objects as wireframes (default: False)

    @type wireframe: C{bool}

    @param lights: Lighting for scene rendered by Camera

    @type lights: C{ILightGroup}

    @param effects: List of optional effects

    @type effects: C{list}

    >>> from miru.components import verifyObject, verifyClass
    >>> from miru.imiru import IPositional
    >>> verifyClass(IPositional, Camera)
    True
    >>> verifyObject(IPositional, Camera(light_group=None))
    True
    >>> verifyClass(ICamera, Camera)
    True
    >>> verifyObject(ICamera, Camera(light_group=None))
    True
    >>> verifyClass(IWorldRenderStage, Camera)
    True
    >>> verifyObject(IWorldRenderStage, Camera(light_group=None))
    True
    """
    implements(ICamera, IWorldRenderStage)
   
    ORBIT_MODE = 1
    ROTATE_MODE = 2
    rotation_mode = ORBIT_MODE
    track_target = None

    def __init__(self, *p, **kw):
        self.wireframe = kw.get('wireframe', False)
        self.objects = self.visible = []
        self.lights = kw.get('lights', LightGroup())
        self.effects = kw.get('effects', [])
        super(Camera, self).__init__(*p, **kw)

    def addobj(self, obj):
        self.objects.append(obj)

    def delobj(self, obj):
        self.objects.remove(obj)

    def render(self, select_pass=0, visible=None, lights=None,
            effect_pass=0, before_render=None):
        if not effect_pass or select_pass:
            for effect in self.effects:
                effect.enable()

        angle = self.angle

        if not select_pass:
            glLoadIdentity()

        p = self.pos
        p = (-p.x, -p.y, -p.z)

        if self.track_target:
            t = self.track_target.pos
            up = (t - self.pos).normalized()
            eux = up.cross(Vector3(0,1,0))
            up = eux.cross(up)
            gluLookAt(self.pos.x, self.pos.y, self.pos.z,
                    t.x, t.y, t.z, up.x, up.y, up.z)
        elif self.rotation_mode == Camera.ORBIT_MODE:
            if not select_pass:
                glTranslatef(*p)
                glRotatef(angle.z, 0, 0, 1)
                glRotatef(angle.y, 0, 1, 0)
                glRotatef(angle.x, 1, 0, 0)
        else:
            if not (select_pass or effect_pass):
                glRotatef(angle.z, 0, 0, 1)
                glRotatef(angle.y, 0, 1, 0)
                glRotatef(angle.x, 1, 0, 0)
                glTranslatef(*p)

        lights = (lights, self.lights)[lights is None]

        if not select_pass:
            if self.wireframe:
                glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
            else:
                glPolygonMode(GL_FRONT, GL_FILL)
            lights.on()

        visible = visible or self.objects

        if not select_pass:
            if effect_pass:
                beforeRender = before_render or (lambda : None)
                beforeRender()
            for v in visible:
                v.draw()
        else:
            for (idx,v) in enumerate(visible):
                glLoadName(idx)
                v.draw()

        if not select_pass:
            lights.off()

        #if not select_pass:
        #    for v in env.debugVisible:
        #        v.draw()

        if self.wireframe and not select_pass:
            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
        
        if not effect_pass or select_pass:
            for effect in self.effects:
                effect.disable()


class DebugView:
    """
    >>> from zope.interface.verify import verifyClass, verifyObject
    >>> verifyClass(IDebuggingRenderStage, DebugView)
    True
    >>> verifyObject(IDebuggingRenderStage, DebugView())
    True
    """
    implements(IDebuggingRenderStage)

    def __init__(self, objects=None):
        self.objects = (objects, [])[objects is None]

    def render(self):
        for o in self.objects:
            o.draw()

    def addobj(self, obj):
        self.objects.append(obj)

    def delobj(self, obj):
        self.objects.remove(obj)

class BlittableView:
    """
    >>> from zope.interface.verify import verifyClass, verifyObject
    >>> verifyClass(IBlittableRenderStage, BlittableView)
    True
    >>> verifyObject(IBlittableRenderStage, BlittableView())
    True
    """
    implements(IBlittableRenderStage)

    def __init__(self):
        self.objects = []

    def _cmp(self, o1, o2):
        return cmp(o1.pos.z, o2.pos.z)

    def _check_zorder(self, moved, pos, delta):
        if moved not in self.objects:
            # XXX this would be unusual?? hrmmm ... remove
            # from positional listeners?
            moved.pos_listeners.remove(self._check_zorder)
            return
        # This is uuuugly - but we want to avoid sorting potentially
        # large lists if we don't have to.
        # We do a custom sort in place as an optimization.
        if delta.z:
            idx = self.objects.index(moved)
            if idx:
                left = self.objects[idx-1]
                while pos.z < left.pos.z:
                    o = self.objects.pop(idx)
                    idx -= 1
                    self.objects.insert(idx, o)
                    if (idx - 1) < 0:
                        return
                    left = self.objects[idx - 1]
            if idx < len(self.objects) - 1:
                right = self.objects[idx+1]
                while pos.z > right.pos.z:
                    o = self.objects.pop(idx)
                    idx += 1
                    self.objects.insert(idx, o)
                    if (idx + 1) > len(self.objects) - 1:
                        return
                    right = self.objects[idx + 1]

    def render(self):
        glPushAttrib(GL_ENABLE_BIT)
        glEnable(GL_BLEND)
        glDisable(GL_DEPTH_TEST)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        glEnable(GL_TEXTURE_2D)
        glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT)

        for o in self.objects:
            o.draw()

        glPopClientAttrib()
        glPopAttrib()
    
    def addobj(self, o):
        self.objects.append(o)
        if self._check_zorder not in o.pos_listeners:
            o.pos_listeners.append(self._check_zorder)
        self.objects.sort(self._cmp)


    def delobj(self, o):
        if self._check_zorder in o.pos_listeners:
            o.pos_listeners.remove(self._check_zorder)
        self.objects.remove(o)

_debugView = DebugView()
components.registerUtility(IDebuggingRenderStage, _debugView)
_blittableView = BlittableView()
components.registerUtility(IBlittableRenderStage, _blittableView)

class LightMixin(object):

    light_no = GL_LIGHT0

    def __init__(self, *p, **kw):
        #self.pos = kw.get('pos', (0.5,0.5,1.0))
        #self.angle = Vector3(0.0, 0.0, 0.0)  # XXX setting the angle has no effect
        self.ambient = kw.get('ambient', (0.5,0.5,0.5,1))
        self.specular = kw.get('specular', (0.9,0.9,0.9,1))
        self.diffuse = kw.get('diffuse', (1,1,1,1))
        super(LightMixin, self).__init__(*p, **kw)

    def _getspecular(self):
        # TODO
        # get value from GL vector - maximize space
        return self._specular

    def _setspecular(self, specular):
        self._specular = specular
        self.lspec = glvec(*specular)

    def _getdiffuse(self):
        return self._diffuse

    def _setdiffuse(self, diffuse):
        self._diffuse = diffuse
        self.ldifs = glvec(*diffuse)
    
    def _getambient(self):
        return self._ambient

    def _setambient(self, ambient):
        self._ambient = ambient
        self.lambient = glvec(*ambient)

    specular = property(_getspecular, _setspecular)
    diffuse = property(_getdiffuse, _setdiffuse)
    ambient = property(_getambient, _setambient)
    

def _update_light_pos(self, pos, _delta):
    self._pos = pos
    self.lpos = glvec(*(tuple(pos) + (self.W,)))

class DirectionalLight(LightMixin, PositionalMixin):

    W = 0.0

    def __init__(self, *p, **kw):
        # change default for pos
        if not kw.has_key('pos'):
            kw['pos'] = (0.5,0.5,1.0)
        super(DirectionalLight, self).__init__(*p, **kw)
        self.pos_listeners.append(_update_light_pos)
        _update_light_pos(self, self.pos, None)

    def on(self):

        lno = self.light_no
        glEnable(lno)

        # Define a simple function to create ctypes arrays of floats:

        glLightfv(lno, GL_POSITION, self.lpos)
        glLightfv(lno, GL_SPECULAR, self.lspec)
        glLightfv(lno, GL_DIFFUSE, self.ldifs)
        glLightfv(lno, GL_AMBIENT, self.lambient)


class PositionalLight(LightMixin, PositionalMixin):
    """A positinal light's pos is the actual position in 3D space coords.
    The attenuation can be adjusted via the three properties kc, kl and kq
    which are constant, linear and quadratic parts.

    Additionaly, a positional light can be turned into a spotlight by
    specifying a spot_cutoff with value in the range [0,90].  By default
    the spot_cutoff is the special value 180 and behaves as a normal
    positional light source.  Other parameters related to spot lights
    (spot_dir, spot_exponent and track_target) will have no effect if
    spot_cutoff is set to 180.

    Related formula:

    attenuation = 1 / (k_c + k_l*d + k_q*d^2)

    k_c = GL_CONSTANT_ATTENUATION
    k_l = GL_LINEAR_ATTENUATION
    k_q = GL_QUADRATIC_ATTENUATION

    @param kc: constant attenuation

    @param kl: linear attenuation

    @param kq: quadratic attenuation

    @param spot_cutoff: spot cut off in range [0,90] or 180 for all directions

    @param spot_exponent: the spot exponent in range [0,128]

    @param spot_dir: the spot direction (unit vector)

    @param track_target: positional object the spot light should
        remain focussed on

    """

    W = 1.0

    _debug = False
    _debugGlobe = None
    _debugTrack = None

    def __init__(self, *p, **kw):
        self.kc = kw.get('kc', 0.0)
        self.kl = kw.get('kl', 0.0)
        self.kq = kw.get('kq', 0.5)
        self.spot_cutoff = kw.get('spot_cutoff', 180.0)
        self.spot_dir = kw.get('spot_dir', (0.0,0.0,-1.0))
        self.spot_exponent = float(kw.get('spot_exponent', 3))
        self.track_target = kw.get('track_target', None)
        # change default for pos
        if not kw.has_key('pos'):
            kw['pos'] = (0.5,1.0,1.0)
        super(PositionalLight, self).__init__(*p, **kw)
        self.pos_listeners.append(_update_light_pos)
        _update_light_pos(self, self.pos, None)
        if options['debug_view'] or kw.get('debug'):
            self.debug = True

    def on(self):

        lno = self.light_no
        glEnable(lno)

        glLightf(lno, GL_CONSTANT_ATTENUATION, self.kc)
        glLightf(lno, GL_LINEAR_ATTENUATION, self.kl)
        glLightf(lno, GL_QUADRATIC_ATTENUATION, self.kq)

        if self.is_spot:
            if self.track_target:
                self._track()
            glLightf(lno, GL_SPOT_CUTOFF, self._spot_cutoff)
            glLightfv(lno, GL_SPOT_DIRECTION, self.lspot_dir)
            glLightfv(lno, GL_SPOT_EXPONENT, self.spot_exponent)

        glLightfv(lno, GL_POSITION, self.lpos)
        glLightfv(lno, GL_SPECULAR, self.lspec)
        glLightfv(lno, GL_DIFFUSE, self.ldifs)
        glLightfv(lno, GL_AMBIENT, self.lambient)

    def _track(self):
        v = (self.track_target.pos - self._pos).normalize()
        self.spot_dir = v

    def _getspotcutoff(self):
        return self._spot_cutoff

    def _setspotcutoff(self, angle):
        assert angle == 180.0 or (angle >= 0.0 and angle <= 90.0)
        self.is_spot = (angle != 180.0)
        self._spot_cutoff = angle

    def _getspotdir(self):
        return self._spot_dir

    def _setspotdir(self, direction):
        self._spot_dir = Vector3(*direction)
        self.lspot_dir = glvec(*direction)

    def _getspotexponent(self):
        return self._spot_exponent

    def _setspotexponent(self, exp):
        self._spot_exponent = (1 * GLfloat)()
        self._spot_exponent[:] = [exp]

    spot_cutoff = property(_getspotcutoff, _setspotcutoff)
    spot_dir = property(_getspotdir, _setspotdir)
    spot_exponent = property(_getspotexponent, _setspotexponent)

    def _setdebug(self, debug):

        try:
            from miru.environment import env
        except ImportError:
            from warnings import warn
            warn('Cannot set debug on light instances until '
                    'environment module is imported')
            return

        from miru import track, mesh
        from miru.environment import env
        self._debug = debug
        if debug:
            if not self._debugGlobe:
                self._debugGlobe = mesh.newObject(
                        mesh._Pin, radius=0.07).withMaterial(
                                ambient=(0,0,0,0.1),
                                diffuse=self.diffuse,
                                specular=(0,0,0,1),
                            ).inRenderStages(IDebuggingRenderStage, IWorldRenderStage)
                self._debugGlobe.pos = self.pos
            env.addobj(self._debugGlobe)
            self._debugTrack = track.PosTrack(self, self._debugGlobe)
            track.PosTrack(self._debugGlobe, self)
        else:
            if self._debugGlobe:
                env.delobj(self._debugGlobe)
                self._debugTrack.untrack()
                try:
                    track = track.gettracks(track=(self._debugTrack, self))
                    track.untrack()
                except:
                    pass

    def _getdebug(self):
        return self._debug

    debug = property(_getdebug, _setdebug)


class LightGroup(UserList):
    """Known issue:

    On slicing the light group list a sublist of copies (vs. same instances)
    is returned.  Current workaround is to only get instances via indexing
    into the light group.

    @param lights: List of lights to add to the group
    """

    def __init__(self, lights=None):
        UserList.__init__(self)
        self.extend(lights or [])
        

    def on(self):
        glPushMatrix()
        glEnable(GL_DEPTH_TEST)
        glPushAttrib(GL_LIGHTING_BIT)
        glEnable(GL_LIGHTING)
        for l in self:
            l.on()
    
    def off(self):
        glPopAttrib()
        #glDisable(GL_DEPTH_TEST)
        glPopMatrix()

    def append(self, light):
        _constrain_lights(len(self) + 1)
        #clone = copy(light)
        #setattr(clone, 'light_no', lightno(len(self)))
        light.light_no = lightno(len(self))
        UserList.append(self, light)

    def extend(self, lights):
        _constrain_lights(len(self) + len(lights))
        lts = []
        for i, l in enumerate(lights):
            #clone = copy(l)
            #setattr(clone, 'light_no', lightno(len(self) + i))
            l.light_no = lightno(len(self) + i)
            lts.append(l)
        UserList.extend(self, lts)


def _camera_on_translate(handle, x, y, dx, dy):
    self = handle

    f = self.camera_move_factor
    if self.axis == self.AXIS_X:
        self.target.pos += (f * dx, 0, 0)
    elif self.axis == self.AXIS_Y:
        self.target.pos += (0, f * dy, 0)
    elif self.axis == self.AXIS_Z:
        d = math.sqrt(dx**2 + dy**2)
        d *= (3,-3)[dy < 0]
        self.target.pos += (0, 0, f * d)
    else:
        self.target.pos += (f * dx, f * dy, 0)

@implementer(ITranslationHandler)
def _camera_on_translate_adpt(_context):
    return _camera_on_translate

registerAdapter(_camera_on_translate_adpt, ICamera, ITranslationHandler)

@implementer(IRotationHandler)
def _camera_on_rotate_adpt(_context):
    from miru import mesh
    return mesh._worldobj_on_rotate

registerAdapter(_camera_on_rotate_adpt, ICamera, IRotationHandler)

