# Copyright (c) 2008 Drew Smathers.
# See LICENSE for details
from euclid import Vector3

from miru.imiru import IPositional
from miru.components import implements

# XXX For our PositionalMixin we need to know if
# the value was updated like v+= (x,y,z) before
# setting
class _Vector3(Vector3):
    _last_value = None
    def __iadd__(self, other):
        self._last_value = tuple(self)
        return super(_Vector3, self).__iadd__(other)
    def __isub__(self, other):
        self._last_value = tuple(self)
        v = Vector3(*other)
        return super(_Vector3, self).__iadd__(-1 * v)
    def __imul__(self, scalar):
        self._last_value = tuple(self)
        return super(_Vector3, self).__imul__(scalar)

def _getX(attr):
    a = '_%s' % attr
    def m(self):
        return getattr(self, a)
    return m

def _setX(attr):
    a = '_%s' % attr
    def m(self, value):
        if value is None:
            return
        if type(value) is tuple:
            value = Vector3(*value)

        constraints = getattr(self, '%s_constraints' % attr)
        for constr in constraints:
            value = constr.constrain(value)

        curr = getattr(self,a)
        if curr._last_value is not None:
            v = Vector3(*(getattr(self, a))._last_value)
            if curr != v:
                setattr(self, a, _Vector3(*tuple(value)))
        else:
            v = getattr(self, a)
            setattr(self, a, _Vector3(*tuple(value)))
        listeners =  getattr(self, '%s_listeners' % attr)
        if v is not None:
            for f in listeners:
                f(self, value, value - v)
    return m

class PositionalMixin(object):
    """Positional object have pos and angle attributes
    which we can easily add listeners to do for immediate
    notification of changes.

    Example usage:

    >>> class Foo(PositionalMixin): pass
    ...
    >>> def myListener(object, pos, delta):
    ...    print 'Called:', tuple(pos), tuple(delta)
    ...
    >>> foo = Foo(pos_listeners=[myListener])
    >>> foo.pos = (1,1,1)
    Called: (1, 1, 1) (1, 1, 1)
    >>> foo.pos += (2,2,2)
    Called: (3, 3, 3) (2, 2, 2)
    """
    implements(IPositional)

    def __init__(self, *p, **kw):
        for a in ('pos', 'angle'):
            if kw.has_key(a):
                setattr(self, '_%s' % a, _Vector3(*tuple(kw[a])))
            else:
                setattr(self, '_%s' % a, _Vector3(0,0,0))
        self.pos_listeners = kw.get('pos_listeners', [])
        self.angle_listeners = kw.get('angle_listeners', [])
        self.pos_constraints = kw.get('pos_constraints', [])
        self.angle_constraints = kw.get('angle_constraints', [])


    _getpos = _getX('pos')
    _setpos = _setX('pos')
    pos = property(_getpos, _setpos)

    _getangle = _getX('angle')
    _setangle = _setX('angle')
    angle = property(_getangle, _setangle)


def pointInRect(x, y, rx, ry, width, height):
    """2D point inside rectangle check
    """
    return (x >= rx) and (x <= rx + width) and\
           (y >= ry) and (y <= ry + height)


def changeCursor(style):
    from miru.environment import env
    win = env.window
    c = win.get_system_mouse_cursor(style)
    win.set_mouse_cursor(c)


options = {
    'debug_view' : False
}


