# Copyright (c) 2008 Drew Smathers.
# See LICENSE for details
from pyglet.gl import *
from pyglet.window import key, mouse
from euclid import Vector3

from miru import imiru

from miru.imiru import IDraggable, ITranslationHandler, IRotationHandler, IDevice
from miru import mesh
from miru import common

from zope.interface import implements

from twisted.python import failure

import ctypes
import sys
import math

class Handle(object):
    implements(IDevice)

    # Transformation constants
    TR_NOOP = 0
    TR_TRANSLATE = 1
    TR_ROTATE = 2
    TR_SCALE = 3

    AXIS_NONE = 0
    AXIS_X = 1
    AXIS_Y = 2
    AXIS_Z = 3

    camera_move_factor = 0.005

    _last_over = None

    def __init__(self, target=None, accessible=False):
        self.target = target
        self.focussed = None
        self.mode = Handle.TR_NOOP
        self.accessible = accessible
        self.axis = Handle.AXIS_NONE
        self._orig_pos = Vector3(0, 0, 0)
        self._orig_angle = Vector3(0, 0, 0)


    def activate(self, window=None):
        from miru.environment import env
        window = window or env.window
        window.push_handlers(self.on_key_press)
        window.push_handlers(self.on_mouse_motion)
        window.push_handlers(self.on_mouse_press)
        window.push_handlers(self.on_mouse_release)
        window.push_handlers(self.on_mouse_drag)
        return self

    def on_key_press(self, symbol, modifiers):
        if self.target is None:
            return
        #if self.mode == Handle.TR_NOOP:
        start = False
        if symbol == key.G and self.mode in (Handle.TR_NOOP, Handle.TR_ROTATE):
            self.mode = Handle.TR_TRANSLATE
            start = True
        elif symbol == key.R and self.mode in (Handle.TR_NOOP, Handle.TR_TRANSLATE):
            self.mode = Handle.TR_ROTATE
            start = True
        elif symbol == key.S:
            from warnings import warn
            warn("Scaling mode is not currently supported")
            self.mode = Handle.TR_SCALE

        if start:
            self._orig_pos = Vector3(*self.target.pos)
            self._orig_angle = Vector3(*self.target.angle)

        if self.mode != Handle.TR_NOOP:
            if symbol == key.X:
                self.axis = (Handle.AXIS_X, Handle.AXIS_NONE)\
                        [self.axis == Handle.AXIS_X]
            elif symbol == key.Y:
                self.axis = (Handle.AXIS_Y, Handle.AXIS_NONE)\
                        [self.axis == Handle.AXIS_Y]
            elif symbol == key.Z:
                self.axis = (Handle.AXIS_Z, Handle.AXIS_NONE)\
                        [self.axis == Handle.AXIS_Z]

    def on_mouse_motion(self, x, y, dx, dy):
        self._move(x, y, dx, dy, drag=self.accessible)

    def _move(self, x, y, dx, dy, button=None, modifiers=None, drag=False):
        if drag:
            if self.target and self.mode == Handle.TR_TRANSLATE:
                try:
                    handler = ITranslationHandler(self.target)
                except TypeError:
                    failure.Failure().printTraceback()
                    return
                return handler(self, x, y, dx, dy)

            elif self.target and self.mode == Handle.TR_ROTATE:
                try:
                    handler = IRotationHandler(self.target)
                except TypeError:
                    failure.Failure().printTraceback()
                    return
                return handler(self, x, y, dx, dy)
        self._check_widgets_device_over(x, y, dx, dy)


    def _check_widgets_device_over(self, x, y, dx, dy):
        from miru.environment import env
        hit = None
        for o in env.osd.widgets:
            try:
                w = imiru.IWidget(o)
            except TypeError:
                pass
            if common.pointInRect(x, y, o.pos.x,
                    o.pos.y, w.width, w.height):
                hit = w
        if hit:
            if hit == self._last_over:
                hit.on_device_hover(self, x, y, dx, dy)
                return
            if self._last_over:
                self._last_over.on_device_exit(self, x, y)
            hit.on_device_enter(self, x, y)
            hit.on_device_hover(self, x, y, dx, dy)
            self._last_over = hit
        else:
            if self._last_over:
                self._last_over.on_device_exit(self, x, y)
                self._last_over = None
            

    def on_mouse_press(self, x, y, button, modifiers):
        from miru.environment import env

        # Middle mouse button allows us to manipulate 
        # the camera orientation
        if button == mouse.MIDDLE:
            self.target = env.camera
            self.mode = Handle.TR_ROTATE
            if modifiers & key.MOD_SHIFT:
                self.mode = Handle.TR_TRANSLATE
            return 
        
        if not (self.target and self.mode):
            return self._select(x, y, button, modifiers)

        if button == mouse.RIGHT:
            self.target.pos = self._orig_pos
            self.target.angle = self._orig_angle
        elif button == mouse.LEFT:
            self.mode = Handle.TR_NOOP
            self.axis = Handle.AXIS_NONE
            common.changeCursor(env.window.CURSOR_DEFAULT)


    def on_mouse_drag(self, x, y, dx, dy, button, modifiers):
        #if not self.accessible:
        self._move(x, y, dx, dy, button, modifiers, drag=(not self.accessible))

    def on_mouse_release(self, x, y, button, modifiers):
        from miru.environment import env
        foc = self.focussed or self.target
        if foc:
            try:
                w = imiru.IReleasable(foc)
                w.on_device_release(self, x, y, button, modifiers)
            except TypeError:
                pass
        if not self.accessible:
            common.changeCursor(env.window.CURSOR_DEFAULT)
            self.target = None


    def _select(self, x, y, button, modifiers):
        from miru.environment import env

        # First see if we clicked a clickable object in the OSD
        t = None
        for c in env.osd.clickable:
            d = c.drawable
            if (x >= c.pos.x and x <= c.pos.x + d.width)\
                and (y >= c.pos.y and y <= c.pos.y + d.height):
                t = c
        if t:
            try:
                widget = imiru.IWidget(t)
                widget.on_device_press(self, x, y, button, modifiers)
            except TypeError:
                failure.Failure().printTraceback()
            return

        # following technique is adapted from 
        # http://www.cse.msu.edu/~cse872/tutorial9.html

        select_buffer = ctypes.cast((100 * GLuint)(), ctypes.POINTER(GLuint))
        glSelectBuffer(100, select_buffer)
       
        w = env.window.width
        h = env.window.height

        viewport = (4 * GLint)()
        glGetIntegerv(GL_VIEWPORT, viewport)
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()

        # rotate the camera first
        angle = env.camera.angle
        glRotatef(angle.z, 0, 0, 1)
        glRotatef(angle.y, 0, 1, 0)
        glRotatef(angle.x, 1, 0, 0)

        gluPickMatrix(x, y, 3, 3, viewport)
        glRenderMode(GL_SELECT)
        gluPerspective(50., w / float(h), 0.1, 1000.)
        glMatrixMode(GL_MODELVIEW)

        glInitNames()
        glPushName(-1)
        env.camera.render(select_pass=1)

        glFlush()
        hits = glRenderMode(GL_RENDER)
        glPopName()

        if hits:
            try:
                m = sys.maxint << 100
                idx = 0
                for i in range(0, 100, 4):
                    if not select_buffer[i]:
                        self.target = env.camera.objects[idx]
                        self.mode = Handle.TR_TRANSLATE
                        common.changeCursor(env.window.CURSOR_HAND)
                        break
                    m = min(select_buffer[i+1], m)
                    if m == select_buffer[i+1]:
                        idx = select_buffer[i+3]
            except IndexError:
                pass
        
        if not self.target:
            common.changeCursor(env.window.CURSOR_DEFAULT)

        env.window.on_resize(w, h)
        

