# Copyright (c) 2008 Drew Smathers.
# See LICENSE for details
"""Abstract LineEditor with modes.  PythonMode give you and
interactive interpreter.  Note: currently this module depends
on twisted.
"""

from pyglet import font
from pyglet.window import key

from twisted.python import failure
from twisted.python.log import StdioOnnaStick

from euclid import Vector3, Vector2

from miru.mesh import Rectangle
from miru.components import implements
from miru.imiru import ILineEditorMode
from miru.environment import env

from itertools import cycle as itercycle
import sys


##########################################################################
### Console and LineEditor utilties
##########################################################################

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO


class PlainTextMode:
    """
    >>> from miru.components import verifyObject, verifyClass
    >>> verifyClass(ILineEditorMode, PlainTextMode)
    True
    >>> verifyObject(ILineEditorMode, PlainTextMode())
    True
    """
    implements(ILineEditorMode)

    def on_line(self, line):
        pass

    def on_key_press(self, symbol, modifiers):
        pass

    def on_text(self, text):
        pass

class PythonMode:
    """
    >>> from miru.components import verifyObject, verifyClass
    >>> verifyClass(ILineEditorMode, PythonMode)
    True
    >>> verifyObject(ILineEditorMode, PythonMode({}, None))
    True
    """
    implements(ILineEditorMode)

    STATE_DOT = 1

    _alpha = range(ord('A'), ord('Z') + 1)
    PY_KEYS = [ eval('key.%s' % chr(a)) for a in _alpha ]
    PY_KEYS += [ eval('key.NUM_%d' % n) for n in range(10) ]
    PY_KEYS.append(key.UNDERSCORE)
    del _alpha

    def __init__(self, ns, lineEditor):
        self._newnames = set()
        self.lineEditor = lineEditor
        self._g = dict(ns)
        self._l = dict(ns)
        #self._l['env'] = env
        #for name in dir(env): # make everything in the environment global
        #    self._g[name] = getattr(env, name)
        # XXX
        self._first_error = True
        self._first_out = True
        self._state = 0
        self._currvar = ''
        self._exp = [] # TAB expansions
        self._nexp = [] # narrowed TAB expansions
        self._last = '' #self._b = ''
        self._typed = ''

    def on_line(self, line):
        # Badness will ensue if we allow other threads to write to stderr or stdout
        # But there could some interesting tricks around this using thread locals ...
        _stdout = sys.stdout
        _stderr = sys.stderr
        sys.stdout = captured_stdout
        sys.stderr = captured_stderr
        ev = b = e = None
        try:
            try:
                ev = eval(line, self._g, self._l)
            except:
                exec(line, self._g, self._l)
            b = sys.stdout.written
            e = sys.stderr.written
            sys.stdout = _stdout
            sys.stderr = _stderr
        except Exception, se:
            sys.stdout = _stdout
            print >> sys.stderr, str(se)
            e = sys.stderr.written
            sys.stderr = _stderr
            fl = failure.Failure()
            fl.printTraceback()
        except RuntimeError, re:
            sys.stdout = _stdout
            print >> sys.stderr, str(re)
            e = sys.stderr.written
            sys.stderr = _stderr
        return (ev, b, e)


    def _expand_tab(self):
        #env = currenv()
        typed = self.lineEditor._lhs_buffer.getvalue().split()[-1]
        cs = list(typed)
        cs.reverse()
        for idx, c in enumerate(cs):
            if c.isalnum() or c == '.':
                continue
            else:
                idx -= 1
                break
        typed = typed[len(typed) - idx - 1:]
        self._typed = typed
        if typed[-1] == '.':
            typed = typed[:-1]
            ctx = None
        else:
            ctx = typed.split('.')[-1]
            typed = '.'.join(typed.split('.')[:-1])
        try:
            self._currvar = eval(typed, self._g, self._l)
            self._exp = [ w for w in dir(self._currvar) ]#if w[0] != '_' ]
        except NameError:
            self._exp = [ w for w in self._l.keys() if w[:len(typed)]==typed]
        #if not self._exp:
        #    # try global names
        #    self._exp = [ w for w in self._g.keys() if w[:len(self._)]
        assert self._exp, 'no expansions!'
        if ctx:
            self._nexp = itercycle([ e for e in self._exp if e[:len(ctx)] == ctx])
        else:
            self._nexp = itercycle([ e for e in self._exp])
            self._state |= PythonMode.STATE_DOT

    def on_key_press(self, symbol, modifiers):
        if symbol == key.TAB:# and (self._state & PythonMode.STATE_DOT):
            if not (self._state & PythonMode.STATE_DOT):
                try:
                    self._expand_tab()
                except:
                    return
            l = self._last
            lb = self.lineEditor._lhs_buffer
            try:
                next = self._nexp.next()
            except StopIteration:
                return
            if next == l:
                return
            if len(l):
                lb.seek(lb.tell() - len(l))
                lb.truncate()
                lb.flush()
            b = self._typed.split('.')[-1]
            lb.write(next[len(b):])
            self.lineEditor._recache()
            self._last = next
        elif symbol in (key.SPACE, key.ENTER):
            self._state &= (~PythonMode.STATE_DOT)
            self._last = ''
        elif symbol in PythonMode.PY_KEYS:
            pass
        else:
            self._state &= (~PythonMode.STATE_DOT)
            self._last = self._b = ''

    def on_text(self, text):
        if text == '.':
            return


class StdioOnnaNotherStick(StdioOnnaStick):

    def __init__(self, *p, **kw):
        StdioOnnaStick.__init__(self, *p, **kw)
        self._written = []

    @property
    def written(self):
        v = '\n'.join(self._written)
        self._written = []
        return v

    def write(self, data):
        d = (self.buf + data).split('\n')
        self.buf = d[-1]
        messages = d[0:-1]
        for message in messages:
            self._written.append(message)

    def writelines(self, lines):
        for line in lines:
            pass
            #self._written.append(line)

captured_stdout = StdioOnnaNotherStick()
captured_stderr = StdioOnnaNotherStick(1)

PYTHON_MODE = 0
TEXT_MODE = 1


def _mk_cursor(pos, ascent):
    x,y = pos
    y = y - 2
    return Rectangle(width=2, height=ascent + 2,
            pos=(x,y,0), color=(1.0,1.0,1.0,1.0))


class Console(object):

    def __init__(self, **kw):
        """Arguments to Console

        @param window: the window this console is attached to
        @param lineEditor: the line editor
        @param font: font to use for rendering text (optional)
        @param leading: leading for text display (optional)
        @param margin: margin between console and window (optional)
        """
        self.lineEditor = kw.get('lineEditor')
        self.lineEditor.console = self
        self.window = w = kw.get('window')
        self.font = kw.get('font', font.load('Arial', 14))
        leading = kw.get('leading', 10)
        self._margin = margin = Vector2(*kw.get('margin', (7,7)))
        self.offset = self.font.ascent + self.font.descent + leading
        self.pos = Vector3(5, w.height - self.offset, 0)
        self._typed = []
        self._current_line = ''
        self._cached = None
        self._mode = PYTHON_MODE
        self._ww, self._wh = width, height = w.width, w.height
        self._cursor = _mk_cursor((0,0), self.font.ascent)
        self._pane = Rectangle(width=width-(margin.x*2), height=height-(margin.y * 2),
                pos=(margin.x,margin.y,0),color=(0.5,0.5,0.5,0.5))
        self.on_resize(width, height)
        # XXX
        w.push_handlers(self.on_resize)

    def _get_margin(self):
        return self._margin

    def _set_margin(self, margin):
        self._margin = margin
        self.on_resize(self._ww, self._wh)

    margin = property(_get_margin, _set_margin)

    def on_resize(self, width, height):
        margin = self.margin
        # XXX shouldn't recreate new rectangle object on every resize
        #self._pane = Rectangle(width=width-(margin.x*2), height=height-(margin.y * 2),
        #        pos=(margin.x,margin.y),color=(0.5,0.5,0.5,0.5))
        if self._pane.size != (width, height):
            self._pane.size = (width-margin.x * 2, height-(margin.y*2))
        oy = height - self._wh
        for line in self._typed:
            line.y += oy
            line.x = margin.x
        self.pos += (0, oy, 0)
        self.pos.x = margin.x
        if self._cached:
            c = self._cached
            c.y += oy
            # FIXME - I believe _text_width has been removed
            self._cursor.pos = (c.x + c._text_width, c.y, 0)
        self._ww, self._wh = width, height


    def ctl_L(self):
        #env = currenv()
        self.pos = Vector3(self.margin.x, self.window.height - self.offset, 0)
        self._typed = []
        self._cached = None

    def enter(self):
        #env = currenv()
        if not self._cached: # XXX
            return
        self.pos -= (0, self.offset, 0)
        self._cached.color = (1, 0.0, 1, 0.8)
        self._typed.append(self._cached)
        if self._mode == PYTHON_MODE:
            self._cached.text = '>>> ' + self._cached.text
        if self._cached.y < (self.margin.y + (self.offset * 2.5)):
            self.ctl_L()
        if self._mode == PYTHON_MODE:
            ev, out, err = self.lineEditor.outerr
            for (head, color, lines) in ((False, (0.75,0.75,0.70,0.9), repr(ev)),
                    (True, (0.0,1.0,0.0,0.9), out),
                    (True, (1.0,0.0,0.0,0.9), err)):
                if lines:
                    for l in lines.split('\n'):
                    #l = line.split('\n')[0]
                        #if not l:
                        #    
                        msg = font.Text(self.font, l)
                        if not head and ev is None:
                            continue
                        #if len(msg.text) > 70:
                        #    msg.text = msg.text[:66] + ' ...'
                        msg.color = color
                        msg.x, msg.y, _z = self.pos
                        self._typed.append(msg)
                        self.pos -= (0, self.offset, 0)
                        if self.pos.y < (self.margin.y + (self.offset * 2.5)):
                            self.ctl_L()

    def _get_current_line(self):
        return self._current_line

    def _set_current_line(self, line):
        self._current_line = line
        self._cached = None

    current_line = property(_get_current_line, _set_current_line)

    def draw(self):
        #env = currenv()
        self._pane.draw()
        if self.current_line != self.lineEditor.text:
            self.current_line = self.lineEditor.text
        for t in self._typed:
            t.draw()
        if not self._cached:
            text = font.Text(self.font, self.current_line)
            text.x, text.y, _z = self.pos
            self._cached = text
        self._cached.text = self.current_line
        self._reset_cursor()
        p = self._mode == PYTHON_MODE
        if p:
            self._cached.text = '>>> ' + self._cached.text
        self._cached.draw()
        if p:
            self._cached.text = self._cached.text[4:]
        self._cursor.draw()

    def _reset_cursor(self):
        c = self._cached
        fmt = ('%s', '>>> %s')[self._mode == PYTHON_MODE]
        t = font.Text(self.font, fmt % self.lineEditor._lhs_buffer.getvalue())
        t.y = -90000
        t.draw()
        self._cursor.pos = (c.x + t._text_width, c.y, 0)
        

class LineEditor(object):
    """Basic line editor
    """

    # XXX
    console = None

    def __init__(self, **kw):
        self.outerr = None
        self.typed = kw.get('typed', [])
        self.history = kw.get('history', [])
        self.typed_limit = kw.get('typed_limit', 100)
        self.history_limit = kw.get('history_limit', 500)
        self._lhs_buffer = StringIO()
        self._rhs_buffer = StringIO()
        self.history_iter = self._bidir_iter()
        self._cached = ''
        self._htrv_dir = 1
        self._tip = None
        self._modes = {
                'text' : PlainTextMode(),
                'python' : PythonMode(globals(), self)}
        self._mode = self._modes['python']

    def enter(self):
        t, h = self.typed, self.history
        lb, rb = self._lhs_buffer, self._rhs_buffer
        line = lb.getvalue() + self._rhs_str()
        t.append(line)
        self.outerr = self._mode.on_line(line)
        if line: h.append(line)
        if len(t) > self.typed_limit: t.pop(0)
        if len(h) > self.history_limit: h.pop(0)
        if len(h) and not h[-1]:
            h.pop()
        self.history_iter = self._bidir_iter()
        for b in lb, rb:
            b.reset()
            b.truncate()
            b.flush()
        self._cached = ''
        self._tip = None
        #yue.store.save(self)


    @property
    def text(self):
        return self._cached

    def _bidir_iter(self):
        """Bidirectional iterator over typed history .... This also
        include the current line that hasn't been appended to history.
        """
        if not self.history:
            return
        h = list(self.history)
        idx = len(h) - 1
        skip = False
        while True:
            if idx >= len(h):
                idx = len(h)
                tip = self._tip
                self._tip = None
                yield tip or ''
            elif idx < 0:
                idx = 0
                yield h[0]
            else:
                yield h[idx]
            idx += self._htrv_dir
    
    def _jump_history(self, direction):
        if not self.history:
            return
        if self._tip is None and direction == 1:
            return
        if self._tip is None: # This is pretty inefficient
            self._tip = self._cached
        lb, rb = self._lhs_buffer, self._rhs_buffer
        self._htrv_dir = direction
        next = self.history_iter.next()
        self._cached = next
        for b in lb, rb:
            b.reset()
            b.truncate()
            b.flush()
        lb.write(next)


    def _rhs_str(self):
        ls = list(self._rhs_buffer.getvalue())
        ls.reverse()
        return ''.join(ls)

    def _recache(self):
        self._cached = self._lhs_buffer.getvalue() + self._rhs_str()

    def on_key_press(self, symbol, modifiers):
        v = self._mode.on_key_press(symbol, modifiers)
        if v:
            print 'mode masked key events'
            return
        if symbol == key.ENTER:
            self.enter()
            self._cached = ''
            # XXX
            self.console.enter()
            self.console.current_line = ''
            #print 'buffer :', str(self)
        elif symbol == key.T and (modifiers & key.MOD_CTRL):
            return
        elif symbol == key.P and (modifiers & key.MOD_CTRL):
            if isinstance(self._mode, PlainTextMode):
                self._mode = self._modes['python']
                self.console._mode = PYTHON_MODE
                print 'switched mode to python'
            else:
                self._mode = self._modes['text']
                self.console._mode = TEXT_MODE
        elif symbol == key.L and 2 == modifiers:
            self.console.ctl_L()
        if symbol != key.ESCAPE:
            return True
        
    def on_text(self, text):
        if text == '\r': # XXX
            return
        self._lhs_buffer.write(text)
        self._recache()
        # XXX
        self.console.current_line = self._cached
        self._mode.on_text(text)
        #print 'buffer :', str(self)
        return True

    def _shift(self, b1, b2):
        b1.seek(b1.tell() - 1)
        c = b1.read(1)
        b1.seek(b1.tell() - 1)
        b1.truncate()
        b2.write(c)

    def _dump(self, b1, b2):
        t = list(b1.getvalue())
        b1.reset()
        b1.truncate()
        t.reverse()
        b2.write(''.join(t))

    def _delete(self, b):
        b.seek(b.tell() - 1) # might not work right with unicode
        b.truncate()
        self._recache()


    def on_text_motion(self, motion):
        if motion == key.MOTION_BACKSPACE:
            self._delete(self._lhs_buffer)
        if motion == key.MOTION_DELETE:
            self._delete(self._rhs_buffer)
        elif motion == key.MOTION_LEFT:
            self._shift(self._lhs_buffer, self._rhs_buffer)
        elif motion == key.MOTION_RIGHT:
            self._shift(self._rhs_buffer, self._lhs_buffer)
        elif motion == key.MOTION_END_OF_LINE:
            self._dump(self._rhs_buffer, self._lhs_buffer)
        elif motion == key.MOTION_BEGINNING_OF_LINE:
            self._dump(self._lhs_buffer, self._rhs_buffer)
        elif motion == key.MOTION_UP:
            self._jump_history(-1)
        elif motion == key.MOTION_DOWN:
            self._jump_history(1)
        return True
           

lineEditor = None
console = None

def show_console():
    global lineEditor, console
    if not lineEditor:
        lineEditor = LineEditor()
        console = Console(window=env.window, lineEditor=lineEditor)
    if console in env.osd:
        print 'console is already visible!'
        return
    env.window.push_handlers(lineEditor.on_key_press, lineEditor.on_text, lineEditor.on_text_motion)
    env.osd.append(console)

def hide_console():
    if console not in env.osd:
        print 'console is already hidden'
        return
    env.window.pop_handlers()
    env.osd.remove(console)

def _show_hide():
    while True:
        yield show_console()
        yield hide_console()
_toggle_console = _show_hide()
def toggle_console():
    _toggle_console.next()



