# Copyright (c) 2008 Drew Smathers.
# See LICENSE for details
"""Effects that can be added to objects in ICamera.render.
"""

#
# Some notes:
#
# basic bledning equation:
#     C_f = (C_5 * S) + (C_d * D)
#     C_f is the final computed color 
#     C_s is the source color
#     C_d it the destination color
#     S source blending factor
#     D destination blending factor
#
# Blending factor can be set with glBlendFunc(S::GLenum, D:GLenum)

from pyglet.gl import *

from miru import mesh
from miru.imiru import IEffect
from miru.components import implements
from miru.environment import env
from miru.camera import LightGroup, PositionalLight, DirectionalLight

import copy

class Blend:
    _gs = globals()
    _ls = locals()
    for n in ('GL_ZERO', 'GL_ONE', 'GL_SRC_COLOR', 'GL_ONE_MINUS_SRC_COLOR',
              'GL_DST_COLOR', 'GL_ONE_MINUS_DST_COLOR', 'GL_SRC_ALPHA',
              'GL_ONE_MINUS_SRC_ALPHA', 'GL_DST_ALPHA',
              'GL_ONE_MINUS_DST_ALPHA', 'GL_CONSTANT_COLOR',
              'GL_CONSTANT_ALPHA', 'GL_ONE_MINUS_CONSTANT_ALPHA','GL_SRC_ALPHA_SATURATE'):
        _ls['_'.join(n.split('_')[1:])] = _gs[n]
    del _gs
    del _ls

    blendEffect = ZERO

class Reflection:
    """Simulate reflection of objects on a surface.  Currently the surface
    must be coplanar with the (X,Z) plane and Y=1.  In the future, we should
    support arbitrary positioning and orientation of the ground surface.

    N.B.:

    * The reflected lights will not mirror changes in global lights if they
    change position etc. during runtime. 
    
    * Global lighting should be set before creating Reflection instances.

    >>> from miru.components import verifyClass, verifyObject
    >>> verifyClass(IEffect, Reflection)
    True
    >>> verifyObject(IEffect, Reflection(lights=[]))
    True
    """
    implements(IEffect)
    objects = []

    def __init__(self, objects=None, ground=None, lights=None, camera=None):
        self.objects = objects or env.objects
        self.camera = camera
        
        ls = lights
        if ls is None: ls = env.camera.lights
        self.lights = LightGroup()
        for l in ls:
            self.lights.append(l)
            l2 = self.lights[-1]
            p = (l2.pos.x, -l2.pos.y, l2.pos.z)
            l2.pos = tuple(p)
        self.ground = ground or mesh.CheckerBoard(even_color=(1,1,1,0.5), odd_color=(0.1,0.1,0.1,0.5))

    def _before_visible(self):
        glScalef(1,-1,1)

    def enable(self):
        camera = self.camera or env.camera
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT)


        glDisable(GL_DEPTH_TEST)
        glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE)

        glEnable(GL_STENCIL_TEST)
        glStencilOp(GL_REPLACE, GL_REPLACE, GL_REPLACE)
        glStencilFunc(GL_ALWAYS, 1, 0xffffffff)

        glEnable(GL_BLEND)
        glBlendFunc(Blend.SRC_ALPHA, Blend.ONE_MINUS_SRC_ALPHA)
        self.ground.draw()

        glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE)
        glEnable(GL_DEPTH_TEST)

        glStencilFunc(GL_EQUAL, 1, 0xffffffff)
        glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP)

        glPushMatrix()
        glCullFace(GL_FRONT)
        glEnable(GL_NORMALIZE)
        camera.render(lights=self.lights, effect_pass=1, before_render=self._before_visible)
        glDisable(GL_NORMALIZE)
        glCullFace(GL_BACK)
        glPopMatrix()

        glDisable(GL_STENCIL_TEST)
        
        glEnable(GL_BLEND)
        glBlendFunc(Blend.SRC_ALPHA, Blend.ONE_MINUS_SRC_ALPHA)
        self.ground.draw()
        glDisable(GL_BLEND)


    def disable(self):
        pass


class Fog(object):
    """
    >>> from miru.components import verifyObject, verifyClass
    >>> verifyClass(IEffect, Fog)
    True
    >>> verifyObject(IEffect, Fog())
    True
    """
    implements(IEffect)

    HINT_NICEST = GL_NICEST
    HINT_FASTEST = GL_FASTEST
    EQ_LINEAR = GL_LINEAR
    EQ_EXP = GL_EXP
    EQ_EXP2 = GL_EXP2

    def __init__(self, color=(0.9,0.9,0.9,1), start=3.0, end=20.0,
            equation=GL_LINEAR, density=1.0, hint=None):
        self.color = color
        self.start = start
        self.end = end
        self.equation = equation
        self.density = density
        self.hint = hint

    def enable(self):
        glEnable(GL_FOG)
        glFogfv(GL_FOG_COLOR, self.color)
        glFogf(GL_FOG_START, self.start)
        glFogf(GL_FOG_END, self.end)
        if self.density is not None:
            glFogf(GL_FOG_DENSITY, self.density)
        if self.hint is not None:
            glHint(GL_FOG_HINT, self.hint)
        glFogi(GL_FOG_MODE, self.equation)

    def disable(self):
        pass
        #glDisable(GL_FOG)

    def _getcolor(self):
        return self._color
    def _setcolor(self, color):
        self._color = (4 * GLfloat)()
        self._color[:] = color
    color = property(_getcolor, _setcolor)


