# Copyright (c) 2008 Drew Smathers.
# See LICENSE for details

"""Global enviroment object for use in applications.

   from miru.enviroment import env
   ...
   for o in env.visible:
      ... do something

@stability: unstable
"""

from miru import components
from miru.camera import Camera, LightGroup, PositionalLight, DirectionalLight
from miru.osd import OSD
from miru.imiru import IEnvironment, IDebuggingDrawable
from miru.imiru import IWorldRenderStage, IRenderStage, IOSDRenderStage
from miru.imiru import IDebuggingRenderStage, IBlittableRenderStage
from miru.components import implements
from miru import components, controls

from twisted.python import log

class Environment(object):
    """ 
    >>> from miru.components import verifyObject, verifyClass
    >>> verifyClass(IEnvironment, Environment)
    True
    >>> verifyObject(IEnvironment, Environment())
    True
    """
    implements(IEnvironment, IRenderStage)

    window = None
    visible = None
    debugVisible = None
    objects = None
    osd = None
    cameras = None
    camera = None
    lights = None
    handle = None

    # Initializer for testing pruposes ...
    def __init__(self, window=None, osd=None, visible=None, objects=None,\
            camera=None, lights=None, handle=None, cameras=None, debugVisible=None):
        self._render_stages = []
        self._render_stage_dict = {}

        self._camera = None
        self.osd = osd
        self.window = window
        self.visible = visible or []
        self.debugVisible = debugVisible or []
        self.objects = objects or []
        self.cameras = cameras or {}
        self.camera = camera
        if not self._camera:
            self._camera = self.cameras.get('main')
        self.lights = lights
        self.handle = handle

    def update(self, otherenv):
        for attr in (k for k in vars(Environment).iterkeys() if k[0] != '_'):
            v = getattr(otherenv, attr)
            if v:
                setattr(self, attr, v)

    def addobj(self, o):
        if o in self.objects:
            return
        stageIfaces = getattr(o, 'renderStages', False)
        self.objects.append(o)
        if not stageIfaces:
            log.err('renderStages not defined for %r' % o)
            return
        for iface in stageIfaces:
            if not self._render_stage_dict.has_key(iface):
                continue
            self._render_stage_dict[iface].addobj(o)
    
    def delobj(self, o):
        if o not in self.objects:
            return
        stageIfaces = getattr(o, 'renderStages', False)
        while o in self.objects:
            self.objects.remove(o)
        if not stageIfaces:
            log.err('renderStages not defined for %r' % o)
            return
        for iface in stageIfaces:
            if not self._render_stage_dict.has_key(iface):
                continue
            self._render_stage_dict[iface].delobj(o)

    def _getcamera(self):
        return self._camera

    def _setcamera(self, camera):
        if camera is None:
            return
        cams = self.cameras.values()
        if camera not in cams:
            self.cameras['main'] = camera
        old = self._camera
        self._camera = camera
        # Register the current camera globally:
        components.registerUtility(IWorldRenderStage,
                self._camera, override=True)
        if old:
            self.setRenderStages(*self._render_ifaces)
    camera = property(_getcamera, _setcamera)


    def _getwindow(self):
        return self._window

    def _setwindow(self, window):
        self._window = window
        if not self._window:
            return
        if self.osd.window != window:
            self.osd.window = window
        if self.handle:
            self.handle.activate(window)

    window = property(_getwindow, _setwindow)

    def _getosd(self):
        return self._osd

    def _setosd(self, osd):
        self._osd = osd
        components.registerUtility(IOSDRenderStage,
                self._osd, override=True)

    osd = property(_getosd, _setosd)

    def setRenderStages(self, *ifaces):
        self._render_stages = []
        self._render_stage_dict = {}
        self._render_ifaces = ifaces
        for iface in ifaces:
            renderStage = components.getUtility(iface)
            self._render_stages.append(renderStage)
            self._render_stage_dict[iface] = renderStage

    def renderStage(self, iface):
        return self._render_stage_dict.get(iface)

    def render(self):
        """Render the environment.
        """
        for stage in self._render_stages:
            stage.render()
    
    def init(self):
        """Optional init method can be called for setting up environment
        with default C{IRenderStage} providers.  You should not this
        method if you initialize your environment with a custom render
        pipeline via C{IEnvironment.setRenderStages} method.
        """
        self.setRenderStages(IWorldRenderStage, IDebuggingRenderStage,
                IBlittableRenderStage, IOSDRenderStage)
        return self

# Global environment singleton
env = Environment(camera = Camera(
        pos=(0,0,4),
        lights=LightGroup([
            PositionalLight(), DirectionalLight()
            ])),
        osd=OSD(),
        handle=controls.Handle()).init()


