# Copyright (c) 2008 Drew Smathers.
# See LICENSE for details

# Drawing primitives

from math import *
from pyglet.gl import *
import pyglet.image

# TODO Get rid of the star import
# TODO Enforce convetion to always import modules only -
# no names from modules - for more coherent namespacing.
from miru.imiru import *
from miru.components import implements
from miru.common import PositionalMixin
from miru.utils import glvec
from euclid import Vector3, Vector2

from twisted.python.components import registerAdapter
from zope.interface import implementer

class Rectangle(PositionalMixin):
    """Rectangle class.

    @param width: The width of the rectangle
    @param height: The height of the rectangle
    @param pos: The position of the rectangle
    @param angle: The orientation of the rectangle

    >>> rect = Rectangle(200, 400, pos=(0,1,2), angle=(0,0,5))
    >>> tuple(rect.pos)
    (0, 1, 2)
    >>> tuple(rect.angle)
    (0, 0, 5)
    >>> rect.width, rect.height
    (200, 400)

    Rectangle implements the IPositional interface:

    >>> from miru.components import verifyClass, verifyObject
    >>> verifyClass(IPositional, Rectangle)
    True
    >>> verifyObject(IPositional, Rectangle(200,50))
    True

    Also, since Rectangle extends PositionalMixin, pos and angle
    are converted to miru.common._Vector3 (monitorable euclid.Vector3
    extensions):

    >>> type(rect.pos), type(rect.angle)
    (<class 'miru.common._Vector3'>, <class 'miru.common._Vector3'>)
    """
    implements(IDrawable)

    def __init__(self, width, height, **kw):
        self.width = width
        self.height = height
        self.color = kw.get('color', (1,1,1,1))
        super(Rectangle, self).__init__(**kw)

    def draw(self):
        # FIXME - Rectangle should be wrapped by Object
        # so this should be rewritten
        (x, y, z)  = tuple(self.pos)
        w, h = self.width, self.height
        glPushMatrix()
        glColor4f(*self.color)
        glRotatef(self.angle.z, 0, 0, 1)
        glRotatef(self.angle.y, 0, 1, 0)
        glRotatef(self.angle.x, 1, 0, 0)
        glBegin(GL_QUADS)
        glVertex3f(x, y, z)
        glVertex3f(x, y + h, z)
        glVertex3f(x + w, y + h, z)
        glVertex3f(x + w, y, z)
        glEnd()
        glPopMatrix()

    def _getsize(self):
        return self.width, self.height

    def _setsize(self, dim):
        self.width = dim[0]
        self.height = dim[1]

    size = property(_getsize, _setsize)


class Sphere(object):
    """A sphere.

    @param radius: the radius of the sphere
    @param slices: slices (passed to gluSphere)
    @param stacks: stacks (passed to glSpehere)
    """
    implements(IDrawable, IWorldObject)
    renderStages = (IWorldRenderStage,)

    _quadric = None

    def __init__(self, radius=0.63, slices=32, stacks=32):
        if not Sphere._quadric:
            Sphere._quadric = gluNewQuadric()
        self.radius = radius
        self.slices = slices
        self.stacks = stacks

    def draw(self):
        gluSphere(self._quadric, self.radius, self.slices, self.stacks)


class _Pin(Sphere):
    """A sphere with a stippled line under it extending along the Y-axis
    towards Y=0. This is really hard-coded for one usecase - debugging positional
    lights.  It is not very useful on a general level.
    """

    renderStages = (IWorldRenderStage)
    end_point = None

    def __init__(self, **kw):
        super(_Pin, self).__init__(**kw)

    def draw(self):
        w = root_object(self)
        m = self._wrapper.material

        glPushAttrib(GL_ENABLE_BIT)
        glDisable(GL_LIGHTING)
        glEnable(GL_BLEND)
        glDisable(GL_DEPTH_TEST)
        glEnable(GL_COLOR_MATERIAL)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        glColor4f(0.83,0.83,0.87,1)
        super(_Pin, self).draw()
        glColor4f(*(m.diffuse[:3] + (0.1,)))
        glBegin(GL_LINES)
        p = list(w.pos)
        p[0] = 0
        p[1] *= -1
        p1 = p[1]
        p[2] = 0
        glVertex3f(*p)
        if not self.end_point:
            p = list(p)
            p[1] = 0
            glVertex3f(*p)
            p[0] += 0.02
        else:
            glVertex3f(*self.end_point)
        glEnd()
        gluSphere(self._quadric, self.radius * 2.2, self.slices, self.stacks)
        glEnable(GL_LINE_STIPPLE)
        glLineStipple(1, 0x33233)
        glBegin(GL_LINES)
        p[1] = p1
        glColor4f(*m.diffuse[:4])
        glVertex3f(*p)
        if not self.end_point:
            p[1] = 0
            glVertex3f(*p)
        else:
            glVertex3f(*self.end_point)
        glColor4f(1,1,1,1)
        glEnd()
        glPopAttrib()


class Points(object):
    implements(IDebuggingDrawable)
    renderStages = (IDebuggingRenderStage,)

    def __init__(self, points=None, color=(1,1,1)):
        self.points = points or []
        self.color = color
        self._len = len(self.points)
        self.list_id = None

    def draw(self):
        #glPushAttrib(GL_ENABLE_BIT)
        #glDisable(GL_LIGHTING)
        if not self.list_id or self._len != len(self.points):
            self.list_id = glGenLists(1)
            glNewList(self.list_id, GL_COMPILE)
            
            glColor3f(*self.color)
            glBegin(GL_POINTS)
            for pt in self.points:
                glVertex3f(*pt)
            glEnd()

            glEndList()
            self._len = len(self.points)
        else:
            glCallList(self.list_id)
        #glPopAttrib(GL_ENABLE_BIT)


class Mesh(object):

    implements(IWorldObject)
    renderStages = (IWorldRenderStage,)
    list_id = None

    def __init__(self, **kw):
        self.vertices = kw.get('vertices')
        self.texcoords = kw.get('texcoords')
        self.normals = kw.get('normals')
        self.vertex_indices = kw.get('vertex_indices')
        self.interleave = kw.get('interleave', True)
        self.normal_indices = kw.get('normal_indices')
        self.texcoord_indices = kw.get('texcoord_indices')
        self.material = kw.get('material', Material())
        self.array = None

    def compile(self):
        # Compile the display list
        self.list_id = glGenLists(1)
        glNewList(self.list_id, GL_COMPILE)
        if self.interleave:
            self._compile_interleaved()
        else:
            self._subcompile()
        glEndList()
        self.makeBoundingBox()
        return self

    def _subcompile(self):
        vertices = self.vertices
        normals = self.normals
        texcoords = self.texcoords
        indices = self.vertex_indices
        normal_indices = self.normal_indices
        texcoord_indices = self.texcoord_indices
        if texcoord_indices and self.material.texture:
            glEnable(self.material.texture.target)
            glBindTexture(self.material.texture.target, self.material.texture.id)
            for (vidxs, tidxs, nidxs) in zip(indices, texcoord_indices, normal_indices):
                glBegin(GL_POLYGON)
                for (idx, tidx, nidx) in zip (vidxs, tidxs, nidxs):
                    glTexCoord2f(*texcoords[tidx])
                    glNormal3f(*normals[nidx])
                    glVertex3f(*vertices[idx])
                glEnd()
            glDisable(self.material.texture.target)
        else:
            for (vidxs, nidxs) in zip(indices, normal_indices):
                glBegin(GL_POLYGON)
                for (idx, nidx) in zip (vidxs, nidxs):
                    glNormal3f(*normals[nidx])
                    glVertex3f(*vertices[idx])
                glEnd()


    def _compile_interleaved(self):
        vertices = self.vertices
        normals = self.normals
        texcoords = self.texcoords
        # First interleave vertices, normals and texcoords into one list
        if not self.array:
            vs = []
            for (vidxs, tidxs, nidxs) in zip(self.vertex_indices,
                    self.texcoord_indices, self.normal_indices):
                for (idx, tidx, nidx) in zip (vidxs, tidxs, nidxs):
                    t, n, v = texcoords[tidx], normals[nidx], vertices[idx]
                    vs.extend(t + n + v)
            self.array = (GLfloat * len(vs))(*vs)
            self.triangles = len(vs) / 8
        
        glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT)
        self.material.enable()
        glInterleavedArrays(GL_T2F_N3F_V3F, 0, self.array)
        glDrawArrays(GL_TRIANGLES, 0, self.triangles)
        self.material.disable()
        glPopClientAttrib()

    def makeBoundingBox(self):
        xs = [p[0] for p in self.vertices]
        ys = [p[1] for p in self.vertices]
        zs = [p[2] for p in self.vertices]
        minX, maxX = min(xs),max(xs)
        minY, maxY = min(ys),max(ys)
        minZ, maxZ = min(zs),max(zs)
        self.boundingBox = BoundingBox(
                (minX, maxX),
                (minY, maxY),
                (minZ, maxZ))
        
    def cleanup(self):
        # Clear some unneeded attributes for garbage collection
        d = vars(self)
        for name in ('vertices', 'normals', 'vertex_indices', 'texcoords',
                'normal_indices', 'texcoord_indices'):
            del d[name]
    
    def draw(self):
        if not self.list_id:
            if self.interleave:
                # TODO change name to _draw_interleaved
                self._compile_interleaved()
            else:
                # TODO change name to _draw
                self._subcompile()
        else:
            glCallList(self.list_id)

        
class MeshGroup(PositionalMixin):
    """@deprecated"""

    def __init__(self, **kw):
        self._pos = kw.get('pos', Vector3(0,0,0))
        self._angle = kw.get('angle', Vector3(0,0,0))
        self._group = kw.pop('group')
        self.objdata = kw.pop('objdata') # This must be set for persistence

    def compile(self):
        self.list_id = glGenLists(1)
        glNewList(self.list_id, GL_COMPILE)
        for pds in self._group:
            pds._subcompile()
        glEndList()
        del vars(self)['_group']
        return self
    
    def draw(self):
        angle = self.angle
        glPushMatrix()
        glTranslatef(*tuple(self.pos))
        glRotatef(angle.z, 0, 0, 1)
        glRotatef(angle.y, 0, 1, 0)
        glRotatef(angle.x, 1, 0, 0)
        glCallList(self.list_id)
        glPopMatrix()

class BoundingBox(PositionalMixin):
    """

    @param xbounds: x bounds (min and max x)
    @type xbounds: tuple
    """

    def __init__(self, xbounds, ybounds, zbounds):
        self.minX, self.maxX = self._orderBounds(xbounds)
        self.minY, self.maxY = self._orderBounds(ybounds)
        self.minZ, self.maxZ = self._orderBounds(zbounds)
    
    def _orderBounds(self, bounds):
        if bounds[0] < bounds[1]:
            return bounds
        else:
            return (bounds[1],bounds[0])
    
    def overlaps(self, other):
        if self.minX > other.maxX or other.minX > self.maxX:
            return False
        if self.minY > other.maxY or other.minY > self.maxY:
            return False
        if self.minZ > other.maxZ or other.minZ > self.maxZ:
            return False
        return True   
    
    def contains(self, other):
        return (self.minX <= other.minX and self.maxX >= other.maxX
            and self.minY <= other.minY and self.maxY >= other.maxY
            and self.minZ <= other.minZ and self.maxZ >= other.maxZ)

    def copy(self):
        return BoundingBox(
                (self.minX, self.maxX),
                (self.minY, self.maxY),
                (self.minZ, self.maxZ))

    def __repr__(self):
        return '<BoundingBox minX=%3.3f maxX=%3.3f'\
                ' minY=%3.3f maxY=%3.3f minZ=%3.3f maxZ=%3.3f %s>'\
                % (self.minX, self.maxX, self.minY, self.maxY, self.minZ, self.maxZ, id(self))

    @property
    def length(self):
        return abs(self.maxX - self.minX)
    
    @property
    def height(self):
        return abs(self.maxY - self.minY)
    
    @property
    def width(self):
        return abs(self.maxZ - self.minZ)

class Object(PositionalMixin):
    """
    >>> from miru.components import verifyObject, verifyClass
    >>> verifyClass(IPositional, Object)
    True
    >>> verifyObject(IPositional, Object())
    True
    """
    implements(IObject, IDrawable)

    transformFunc = None

    def __init__(self, drawable=None, **kw):
        self.drawable = drawable
        if self.drawable:
            self.drawable._wrapper = self
        # experimental part of this interface
        self.material = kw.get('material', None)
        super(Object, self).__init__(self, **kw)
        self.renderStages = getattr(self.drawable, 'renderStages', ())

        self.pushMatrix = glPushMatrix
        self.popMatrix = glPopMatrix
        self._bbox = None

    def draw(self):
        if self.drawable is None:
            return
        if self.transformFunc:
            self.transformFunc(self)
            self.drawable.draw()
            return
        angle = self.angle
        if not getattr(self.drawable, 'skipTransform', False):
            glPushMatrix()
            glTranslatef(*tuple(self.pos))
            glRotatef(angle.z, 0, 0, 1)
            glRotatef(angle.y, 0, 1, 0)
            glRotatef(angle.x, 1, 0, 0)
            if self.material:
                self.material.enable()
                self.drawable.draw()
                self.material.disable()
            else:
                self.drawable.draw()
            glPopMatrix()
        else:
            if self.material:
                self.material.enable()
                self.drawable.draw()
                self.material.disable()
            else:
                self.drawable.draw()


    @property
    def boundingBox(self):
        """The bounding box of the object with abosolute coordinates
        or None if the wrapped drawable does not have a bounding box.
        """
        if hasattr(self.drawable, 'boundingBox'):
            bb = self.drawable.boundingBox
            if not self._bbox:
                self._bbox = bb.copy()
            self._bbox.minX = self.pos.x + bb.minX
            self._bbox.minY = self.pos.y + bb.minY
            self._bbox.minZ = self.pos.z + bb.minZ
            self._bbox.maxX = self.pos.x + bb.maxX
            self._bbox.maxY = self.pos.y + bb.maxY
            self._bbox.maxZ = self.pos.z + bb.maxZ
        return self._bbox
            

    def withMaterial(self, *p, **kw):
        """
        convenience method for chaining with newObject function
       
            obj = newObject(Foo, ...).withMaterial(mat)

        OR

            obj = newObject(Foo, ...).withMaterial(ambient=...)
        """
        if p and isinstance(p[0], Material):
            self.material = p[0]
        else:
            self.material = Material(*p, **kw)
        return self


    def withTransformFunction(self, transformFunc):
        self.transformFunc = transformFunc
        return self

    def inRenderStages(self, *renderStages):
        self.renderStages = tuple(renderStages)
        return self

    def __repr__(self):
        return '<miru.mesh.Object at %x %r>' % (id(self), self.drawable)

# Convenience object factory
def newObject(klass, *p, **kw):
    original = klass(*p, **kw)
    obj = Object(original)
    return obj



class _GroupDrawable:
    implements(IDrawable, IWorldObject)
    renderStages = (IWorldRenderStage,)

    def __init__(self, objects=None, list_id=None):
        if objects:
            self.objects = objects
            self.draw = self._draw_objects
        else:
            self.list_id = list_id
            self.draw = self._draw_list
        try:
            self.boundingBox = self.objects[0].boundingBox
        except:
            print 'Failed computing bounding box: %r' % self

    def _draw_list(self):
        glCallList(self.list_id)

    def _draw_objects(self):
        for o in self.objects:
            o.draw()

class Group(Object):

    _debug = False

    def __init__(self, *objects, **args):
        """An object grouping.

        @param objects: The objects to group

        @param compile: Set to True for objects that don't move relative
                        to the parent so display list can be used.  Use with
                        caution as grouped objects might have their own
                        display lists - and you cannot nest display lists.
        """
        self.objects = objects
        for o in self.objects:
            o._wrapper = self
        self.drawable = None
        compile = args.get('compile', False)
        if compile:
            list_id = self._compile(objects)
            drawable = _GroupDrawable(list_id=list_id)
        else:
            drawable = _GroupDrawable(objects=objects)
        args['drawable'] = drawable
        super(Group, self).__init__(**args)

    # TODO abstract this as a helper function
    def _compile(self, objects):
        list_id = glGenLists(1)
        glNewList(list_id, GL_COMPILE)
        for idx,obj in enumerate(objects):
            if self._debug:
                print 'compile step %4.4d: %r' % (idx, obj)
            obj.draw()
        glEndList()
        return list_id
         
      

# ------------------------------------------------------------------------
# Conversion functions for data parsed from OBJ file

_loaded_meshes = {}

def objmat2mat(mat):
    return Material(ambient=mat.ambient, diffuse=mat.diffuse,
            specular=mat.specular, shininess=mat.shininess, alpha=mat.alpha,
            teximg_filename=mat.teximg_filename)

def obj2mesh(obj, data=None, compile=True, interleave=False):
    """Convert ObjObject to Mesh
    """
    material = Material()
    if obj.material:
        material = objmat2mat(obj.material)
    pds = Mesh(vertices=obj.vertices, normals=obj.normals,
            vertex_indices=obj.vertex_indices, normal_indices=obj.normal_indices,
            texcoord_indices=obj.texcoord_indices,
            texcoords=obj.texcoords, objdata=data,
            material=material, interleave=interleave)
    if compile:
        return pds.compile()
    return pds

def objs2group(d, data):
    """Convert ObjObject dict into MeshGroup

    @deprecated Use Group instead (this is pretty much untested)
    """
    group = []
    for o in d.values():
        group.append(obj2mesh(o, compile=False))
    return MeshGroup(group=group, objdata=data).compile()


def loadObj(filename, flip_normals=False, interleave=True, compile=True):
    from miru.tools import obj
    mesh = _loaded_meshes.get((filename, flip_normals, interleave, compile))
    if mesh is not None:
        return Object(mesh)
    parser = obj.ObjParser(filename, flip_normals=flip_normals)
    f,d,data = parser.parse()
    mesh = obj2mesh(f, data, interleave=interleave, compile=compile)
    _loaded_meshes[(filename, flip_normals, interleave, compile)] = mesh
    return Object(mesh)

# ---------------------------------------------------------------------- 

class Material(object):

    def __init__(self, **kw):
        ambient = kw.get('ambient', (1.0,1.0,1.0))
        diffuse = kw.get('diffuse', (1.0,1.0,1.0))
        self.specular = kw.get('specular', (0.5,0.5,0.5,1))
        self.shininess = kw.get('shininess', 15)
        self.teximg_filename = kw.get('teximg_filename', None)
        self.alpha = a = kw.get('alpha', 0.5)
        self.ambient = tuple(ambient) + (a,)
        self.diffuse = tuple(diffuse) + (a,)
        self.texture = None
        if self.teximg_filename:
            img = pyglet.image.load(self.teximg_filename)
            self.texture = img.texture
            # TODO
            # Currently the texture only gets mapped correctly if dims are power of 2.
            # My attempt to autmoate this as follows results in TypeError.
            
        
    def enable(self):
        glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_LIGHTING_BIT)
        glCullFace(GL_BACK)
        if not self.texture:
            glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, glvec(*self.ambient))
            glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, glvec(*self.specular))
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, glvec(*self.diffuse))
            glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, self.shininess)
        if self.texture:
            glEnable(self.texture.target)
            glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, glvec(*self.ambient))
            # XXX Specular on the textured mesh right now looks really bad ...
            #glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, glvec(*self.specular))
            glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, glvec(*self.diffuse))
            glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, self.shininess)
            glBindTexture(self.texture.target, self.texture.id)

    def disable(self):
        """Kind of a bad name yes - unapply texture or whatever.
        """
        glPopAttrib()


class Color(Material):

    def __init__(self, r, g, b, a=1.0):
        self.color = (r,g,b,a)

    def enable(self):
        glColor4f(*self.color)

    def disable(self):
        pass


class CartesianPlane:
    implements(IDebuggingDrawable)
    renderStages = (IDebuggingRenderStage,)

    def __init__(self, axes=None, magnitude=50, step=2,\
            color=(1,1,1,1), stipple_pattern=0x55555, stipple_factor=2):
        self.axes = axes or (1,0,1) # xz is default
        self.color = color
        self._compile(magnitude, step, stipple_pattern, stipple_factor)

    def _compile(self, mag, step, stipple_pattern, stipple_factor):
        # Compile the display list
        self.list_id = glGenLists(1)
        glNewList(self.list_id, GL_COMPILE)
        glPushAttrib(GL_ENABLE_BIT)
        if stipple_pattern is not None:
            glEnable(GL_LINE_STIPPLE)
            glLineStipple(stipple_factor, stipple_pattern)
        #glDisable(GL_LIGHTING)
        glColor4f(*self.color)
        if self.axes[0]:
            glBegin(GL_LINES)
            for i in range(-mag, mag + step, step):
                glVertex3f(i, self.axes[1] * (-mag), self.axes[2] * (-mag))
                glVertex3f(i, self.axes[1] * mag, self.axes[2] * mag)
            glEnd()
        if self.axes[1]:
            glBegin(GL_LINES)
            for i in range(-mag, mag + step, step):
                glVertex3f(self.axes[0] * (-mag), i, self.axes[2] * (-mag))
                glVertex3f(self.axes[0] * mag, i, self.axes[2] * mag)
            glEnd()
        if self.axes[2]:
            glBegin(GL_LINES)
            for i in range(-mag, mag + step, step):
                glVertex3f(self.axes[0] * (-mag), self.axes[1] * (-mag), i)
                glVertex3f(self.axes[0] * mag, self.axes[1] * mag, i)
            glEnd()
        glColor4f(1,1,1,1)
        glPopAttrib()
        glEndList()

    def draw(self):
        glCallList(self.list_id)

CoordPlane = CartesianPlane

class CheckerBoard(object):
    implements(IDrawable)
    renderStages = (IWorldRenderStage,)
    
    # TODO - right now just black and white quads, so support
    #        other color patterns - and alpha as well
    def __init__(self, magnitude=50, step=2, odd_color=(0.3,0.3,0.2,1),
            even_color=(0.8,0.9,0.9,1)):
        self._odd_color = odd_color
        self._even_color = even_color
        self.magnitude = magnitude
        self.step = step
        self._compile()

    def _compile(self):
        mag = self.magnitude
        step = self.step

        # Compile the display list
        self.list_id = glGenLists(1)
        glNewList(self.list_id, GL_COMPILE)

        specRef = (4 * GLfloat)()
        specRef[:] = [1.,1.,1.,1.]
        
        glPushAttrib(GL_ENABLE_BIT)
        glEnable(GL_COLOR_MATERIAL)
        glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE)
        
        colors = [self.even_color, self.odd_color]

        glTranslatef(-mag, 0, -mag)
        for k in (0,1):
            s = k
            glColor4f(*colors[k])
            for j in range(mag):
                s += 1
                s %= 2
                j2 = j * step
                for i in range(s, mag, 2):
                    i2 = i * step
                    glBegin(GL_QUADS)
                    glNormal3f(0, 1, 0)
                    glVertex3f(i2, 0, j2)
                    glVertex3f(i2, 0, j2 + step)
                    glVertex3f(i2 + step, 0 ,j2 + step)
                    glVertex3f(i2 + step, 0, j2)
                    glEnd()
        """
        istrip = -mag
        bounce = 0
        while istrip <= mag:
            glBegin(GL_TRIANGLE_STRIP)
            irun = mag
            while irun >= -mag:
                clr = (self.odd_color, self.even_color)[bounce % 2]
                glColor4f(*clr)
                glVertex3f(istrip,0,irun)
                glVertex3f(istrip + step,0,irun)
                irun -= step
                bounce += 1
            glEnd()
            istrip += step
        """
        glPopAttrib()
        glEndList()

    def _getoddcolor(self):
        return self._odd_color
    def _setoddcolor(self, c):
        self._odd_color = c
        self._compile()
    odd_color = property(_getoddcolor, _setoddcolor)

    def _getevencolor(self):
        return self._even_color
    def _setevencolor(self, c):
        self._even_color = c
        self._compile()
    even_color = property(_getevencolor, _setevencolor)

    def draw(self):
        glCallList(self.list_id)


# Figure out a better way to use this ... ?
class ImageWrapper(object):
    implements(IWorldObject)
    renderStages = (IBlittableRenderStage,)
    image_path = None
    skipTransform = True
    _wrapper = None

    def __init__(self, image, scale=0.01):
        self.image = image
        self.scale = scale
        #self.compile()

    def draw(self):
        pos = self._wrapper.pos
        self.image.blit(pos.x, pos.y, pos.z,
                width=self.image.width*self.scale,
                height=self.image.height*self.scale)

@implementer(ITranslationHandler)
def _worldobj_on_translate(handle, x, y, dx, dy):
    self = handle
    target = self.target

    viewport = (GLint * 4)()
    mvmatrix = (GLdouble * 16)()
    projmatrix = (GLdouble * 16)()
    glGetIntegerv(GL_VIEWPORT, viewport)
    glGetDoublev(GL_MODELVIEW_MATRIX, mvmatrix)
    glGetDoublev(GL_PROJECTION_MATRIX, projmatrix)
    wx = GLdouble()
    wy = GLdouble()
    wz = GLdouble()

    sz = GLdouble()
    ex = self.target.pos.x
    ey = self.target.pos.y
    ez = self.target.pos.z
    gluProject(ex, ey, ez, mvmatrix, projmatrix,
            viewport, wx, wy, sz)
    gluUnProject(x, y, sz,
            mvmatrix, projmatrix, viewport, wx, wy, wz)

    if self.axis == self.AXIS_X:
        #target.pos.x = wx.value
        #target.pos.z = wz.value
        target.pos = (wx.value, wz.value, target.pos.z)
    elif self.axis == self.AXIS_Y:
        #target.pos.y = wy.value
        #target.pos.z = wz.value
        target.pos = (target.pos.x, wy.value, wz.value)
    elif self.axis == self.AXIS_Z:
        tz = dy * 0.01
        tz += (dx * 0.01)
        target.pos += (0, 0, tz)
    else:
        #target.pos.x = wx.value
        #target.pos.y = wy.value
        #target.pos.z = wz.value
        target.pos = (wx.value, wy.value, wz.value)

def _worldobj_on_translate_adpt(_context):
    return _worldobj_on_translate
#registerAdapter(_worldobj_on_translate_adpt, IWorldObject, ITranslationHandler)

def _object_on_translate_adpt(obj):
    return ITranslationHandler(obj.drawable)
#registerAdapter(_object_on_translate_adpt, IObject, ITranslationHandler)

@implementer(IRotationHandler)
def _worldobj_on_rotate(handle, x, y, dx, dy):
    self = handle
    if self.axis == self.AXIS_Y:
        tx = dy * 0.35
        self.target.angle +=  (tx, 0, 0)
        self.target.angle.x %= 360.
    elif self.axis == self.AXIS_X:
        ty = dx * 0.35
        self.target.angle += (0, ty, 0)
        self.target.angle.y %= 360.
    elif self.axis == self.AXIS_Z:
        tz = (dy + dx) * 0.35
        self.target.angle += (0, 0, tz)
        self.target.angle.z %= 360.
    else:
        tx, ty = dx * 0.35, dy * 0.35
        self.target.angle += (ty, tx, 0)
        self.target.angle.x %= 360.
        self.target.angle.y %= 360.




def _worldobj_on_rotate_adpt(_context):
    return _worldobj_on_rotate

#registerAdapter(_worldobj_on_rotate_adpt, IWorldObject, IRotationHandler)


# Register adapters for objects

def _obj_adpt_factory(iface):
    def adpt(context):
        return iface(context.drawable)
    return adpt


for iface in (IRotationHandler, IWorldObject,IPressable, IReleasable, IEnterable, IExitable, IWidget):
    a = _obj_adpt_factory(iface)
    registerAdapter(a, IObject, iface)

# XXX - we monkey patch pyglet.image.Texture because it's more
# efficient to only enable the texture mode etc. once for the
# entire render stage.

_pyglet_texture_blitMethod = None

def _patched_blit(self, x, y, z=0, width=None, height=None):
    t = self.tex_coords
    x1 = x - self.anchor_x
    y1 = y - self.anchor_y
    x2 = x1 + (width is None and self.width or width)
    y2 = y1 + (height is None and self.height or height)
    array = (GLfloat * 32)(
         t[0],  t[1],  t[2],  1.,
         x1,    y1,    z,     1.,
         t[3],  t[4],  t[5],  1.,
         x2,    y1,    z,     1.,
         t[6],  t[7],  t[8],  1.,
         x2,    y2,    z,     1.,
         t[9],  t[10], t[11], 1.,
         x1,    y2,    z,     1.)
    glBindTexture(GL_TEXTURE_2D, self.id)
    glInterleavedArrays(GL_T4F_V4F, 0, array)
    glDrawArrays(GL_QUADS, 0, 4)

def _monkeypatchTexture():
    global _pyglet_texture_blitMethod
    from pyglet import image
    _pyglet_texture_blitMethod = image.Texture.blit
    image.Texture.blit = _patched_blit
_monkeypatchTexture()

def unmonkeypatch():
    """Revert monkeypatch on pyglet.image.Texture.
    """
    image.Texture.blit = _pyglet_texture_blitMethod


def root_object(drawable):
    w = drawable._wrapper
    for i in xrange(100):
        w2 = getattr(w, '_wrapper', None)
        if w2 is None:
            return w
        w = w2


