# Copyright (c) 2008 Drew Smathers.
# See LICENSE for details
"""On screen display
"""
from pyglet.gl import *

from miru.components import implements
from miru.imiru import IDraggable, IOSDRenderStage
from UserList import UserList

class OSD(object):
    """The OSD is a container for objects that can be rendered in 2D
    to overlay the 3D sceen.  To add objects:

        osd = OSD(window)
        osd.addobj(thingy)

    The OSD object's on_resize method should be pushed on the event
    stack for handling resize events:

        event.push_handlers('on_resize', osd.on_resize)

    """
    implements(IOSDRenderStage)

    def __init__(self, window=None, objects=None):
        self.objects = (objects, [])[objects is None]
        self.window = window
        # deprecated - widgets make more sense as a generalization than `clickable'
        self.clickable = []
        self.widgets = self.clickable
        
       
    def _getwindow(self):
        return self._window

    def _setwindow(self, window):
        if window is None:
            self._window = window
            return
        self._window = window
        self.width = window.width
        self.height = window.height
        window.push_handlers(self.on_resize)

    window = property(_getwindow, _setwindow)

    def on_resize(self, width, height):
        self.width = width
        self.height = height

    def set_2d(self, near, far):
        glMatrixMode(GL_PROJECTION)
        glPushMatrix()
        glLoadIdentity()
        glOrtho(0, self.width, 0, self.height, near, far)
        glMatrixMode(GL_MODELVIEW)
        glPushMatrix()
        glLoadIdentity()

    def unset_2d(self):
        glPopMatrix()
        glMatrixMode(GL_PROJECTION)
        glPopMatrix()
        glMatrixMode(GL_MODELVIEW)

    def render(self):
        """Draw contained objects.
        """
        if not self.objects:
            return

        r = 32
        self.set_2d(0, r*4)
        glPushAttrib(GL_ENABLE_BIT)
        glEnable(GL_BLEND)
        glDisable(GL_DEPTH_TEST)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        
        for v in self.objects:
            v.draw()
        glColor4f(1,1,1,1)
        glPopAttrib()
        self.unset_2d()

    # TODO For backwards compat - remove later
    draw = render

    def remove(self, obj):
        self.objects.remove(obj)
        if obj in self.clickable:
            self.clickable.remove(obj)

    def pop(self, index=-1):
        obj = self.objects.pop(index)
        if obj in self.clickable:
            self.clickable.remove(obj)
        return obj

    def __iter__(self):
        return iter(self.objects)

    def addobj(self, obj):
        from miru import mesh
        self.objects.append(obj)
        if isinstance(obj, mesh.Object) and IDraggable.providedBy(obj.drawable):
            self.clickable.append(obj)

    def delobj(self, obj):
        self.remove(obj)

    # TODO For backwards compat - remove later
    append = addobj

