# Copyright (c) 2008 Drew Smathers.
# See LICENSE for details

"""Experimental shader functions.

Building and using a shader fuction:

    from miru.shader import *
    vs = VertexShader(data)
    fs = FragmentShader(data)
    program = createProgram(vs, fs)

    ...


@stability: unstable
"""

from pyglet.gl import *
import ctypes as c

from miru import imiru

from zope.interface import implements

def _shaderSource(text, shader_type=GL_VERTEX_SHADER):
    buff = c.create_string_buffer(text)
    c_text = c.cast(c.pointer(c.pointer(buff)), c.POINTER(c.POINTER(GLchar)))
    shader = glCreateShader(shader_type)
    glShaderSource(shader, 1, c_text, None)
    return shader


class GLSLException(Exception):
    """Thrown if shader compilation, linking or validation does not
    succeed
    """

def _infolog(handle, log_type='shader'):
    f = { 'shader' : glGetShaderInfoLog, 'program' : glGetProgramInfoLog }[log_type]
    log = (4096 * GLchar)()
    f(handle, 4096, None, log)
    return log.value

class ShaderMixin:
    """Base shader class.  Do not instantiate me directly.
    """
    shader_type = None

    def __init__(self, source):
        self.shader_no = _shaderSource(source, shader_type=self.shader_type)
        glCompileShader(self.shader_no)
        self.status = GLint()
        glGetShaderiv(self.shader_no, GL_COMPILE_STATUS, self.status)
        if not self.status:
            raise GLSLException(_infolog(self.shader_no))

class VertexShader(ShaderMixin):
    """Vertex shader.  Initialize with vertex shader source data (.vs)
    """
    shader_type = GL_VERTEX_SHADER

class FragmentShader(ShaderMixin):
    """Fragment shader. Initialize with fragment shader source data (.fs)
    """
    shader_type = GL_FRAGMENT_SHADER


_programs = {}

class Program:
    """The shader program.
    """

    link_status = GLint(-1)
    validation_status = GLint(-1)

    def __init__(self):
        self.program_no = glCreateProgram()
        _programs[self.program_no] = self

    def attach(self, shader):
        glAttachShader(self.program_no, shader.shader_no)

    def detach(self, shader):
        glDetachShader(self.program_no, shader.shader_no)

    def link(self):
        glLinkProgram(self.program_no)
        glGetProgramiv(self.program_no, GL_LINK_STATUS, self.link_status)
        if not self.link_status:
            raise GLSLException(_infolog(self.program_no, log_type='program'))

    def validate(self):
        glValidateProgram(self.program_no)
        self.validation_status = GLint()
        glGetProgramiv(self.program_no, GL_VALIDATE_STATUS, self.validation_status)
        if not self.validation_status:
            raise GLSLException(_infolog(self.program_no, log_type='program'))

    def enable(self):
        glUseProgram(self.program_no)

def createProgram(*shaders):
    """Create a Program instance with given shader objects.
    
    @raises GLSLException: If linkage or validation of program fails.
    """
    program = Program()
    for shader in shaders:
        program.attach(shader)
    program.link()
    program.validate()
    return program

def disableShaders():
    """Disable shader programs.  Equivalent to glUseProgram(0)
    """
    glUseProgram(0)

def currentProgram():
    """Get the current enabled program.
    """
    pno = GLint()
    glGetIntegerv(GL_CURRENT_PROGRAM, pno)
    return _programs.get(pno.value)


class ShaderView:
    """
    >>> from zope.interface.verify import verifyClass, verifyObject
    >>> verifyClass(imiru.IShaderRenderStage, ShaderView)
    True
    >>> verifyObject(imiru.IShaderRenderStage, ShaderView())
    True
    """
    implements(imiru.IShaderRenderStage)

    def __init__(self, program=None, objects=None, camera=None):
        self.program = program
        self.camera = camera
        self.objects = (objects, [])[objects is None]

    def render(self):
        self.program.enable()
        self.camera.render(visible=self.objects)
        disableShaders()
    
    def addobj(self, obj):
        self.objects.append(obj)

    def delobj(self, obj):
        self.objects.remove(obj)

