# Copyright (c) 2008 Drew Smathers.
# See LICENSE for details
from miru.test import unittest, path, require_window, get_window

from miru.environment import Environment
from miru import components, mesh, imiru, camera, osd
from zope.interface import implements

import os

require_window()

class EnvironmentTest(unittest.TestCase):
    # The point of the environment object is to expose
    # a global singleton Environment instance env that
    # can be read a different points in the application.

    # This will simulate changes to singleton via calls
    # to replace.
    
    def test_update(self):

        w = object()
        e = Environment(
                visible=['a','b','c'],
                osd=[1,2,3],
                objects=[4,5,6])

        e2 = Environment(
                visible=['d','e','f'],
                osd=[4,5,6])

        e.update(e2)
        self.assertEquals(['d','e','f'], e.visible)
        self.assertEquals([4,5,6], e.osd)
        self.assertEquals([4,5,6], e.objects)

    def testCameraMapping(self):

        camera = object()
        e = Environment(camera=camera)
        self.assertEquals(e.camera, camera)
        self.assertEquals(e.cameras['main'], camera)
        self.assertEquals(1, len(e.cameras.values()))

        camera1 = object()
        camera2 = object()
        cameras = { 'main' : camera1, 'birdseye' : camera2 }
        e = Environment(cameras=cameras)
        self.assertEquals(e.camera, camera1)
        self.assertEquals(e.cameras['main'], camera1)
        self.assertEquals(e.cameras['birdseye'], camera2)
        self.assertEquals(2, len(e.cameras.values()))

        # note that the behavior is undefined if you pass a dictionary with 
        # no main entry and then no default single camera.


    def test_addobj(self):
        o = mesh.loadObj(path('cube.obj'))
        e = Environment()
        self.assertEquals(0, len(e.objects))
        e.addobj(o)
        self.assertEquals(1, len(e.objects))


    def testSettingWindowSideEffects(self):
        o = osd.OSD()
        self.assertNot(o.window)
        e = Environment(osd=o)
        e.window = get_window()
        self.assert_(o.window)

    def testCameraSetting(self):
        # FIXME - this test can leave the suite in a bad state
        from miru.environment import env
        c = env.camera
        self.assertEquals(c, env.renderStage(imiru.IWorldRenderStage))
        cam = camera.Camera()
        env.camera = cam
        self.assertEquals(cam, env.renderStage(imiru.IWorldRenderStage))
        u = components.getUtility(imiru.IWorldRenderStage)
        self.assertEquals(cam, u)
        self.assert_(cam in env._render_stages)
        env.camera = c


class IBlingStage(imiru.IRenderStage):
    pass

class IFoggyStage(imiru.IRenderStage):
    pass

class IUnknownStage(imiru.IRenderStage):
    pass

class RenderStage:
    def __init__(self):
        self.objects = []
        self.render_ct = 0
    def addobj(self, obj):
        self.objects.append(obj)
    def render(self):
        self.render_ct += 1

class BlingStage(RenderStage):
    implements(IBlingStage)
blingStage = BlingStage()
components.registerUtility(IBlingStage, blingStage)

class FoggyStage(RenderStage):
    implements(IFoggyStage)
foggyStage = FoggyStage()
components.registerUtility(IFoggyStage, foggyStage)

class UnknownStage(RenderStage):
    implements(IUnknownStage)
unknownStage = UnknownStage()
components.registerUtility(IUnknownStage, unknownStage)

class Ice(object):
    renderStages = [IBlingStage]

class GasMask(object):
    renderStages = [IFoggyStage]

class GemStone(object):
    renderStages = [IBlingStage, IFoggyStage]

class Martian(object):
    renderStages = [IUnknownStage]

class HalfStranger(object):
    renderStages = [IUnknownStage, IFoggyStage]

class RenderStageTest(unittest.TestCase):

    _camera = None

    def setUp(self):
        try:
            self._camera = components.getUtility(imiru.IWorldRenderStage)
            components.destroyUtility(imiru.IWorldRenderStage)
        except:
            pass

    def tearDown(self):
        if self._camera:
            components.registerUtility(imiru.IWorldRenderStage,
                    self._camera, override=True)

    def testRenderStages(self):
        env = Environment()
        env.setRenderStages(IFoggyStage, IBlingStage, IUnknownStage)
        self.assertEquals(foggyStage, env.renderStage(IFoggyStage))
        self.assertEquals(blingStage, env.renderStage(IBlingStage))
        self.assertEquals(unknownStage, env.renderStage(IUnknownStage))


    def testDefaultWorldRenderStage(self):
        env = Environment()
        cam = camera.Camera(visible=[], lights=[])
        env.camera = cam
        o = components.getUtility(imiru.IWorldRenderStage)
        self.assertEquals(cam, o)

