from miru.test import unittest
from miru import imiru
from miru import mesh
from euclid import Vector3

from zope.interface import implements, Interface

class FooBar:
    implements(imiru.IDrawable)
    renderStages = (imiru.IWorldRenderStage, imiru.IOSDRenderStage)


class IFancyVertexShader(Interface):
    pass

class IFancyFragmentShader(Interface):
    pass


class TestMaterial(mesh.Material):
    def __init__(self, name, color=(1,2,3,4)):
        self.name = name
        self.color = color

class ObjectsTest(unittest.TestCase):

    def testBorrowsRenderStageInterfaces(self):
        o = mesh.newObject(FooBar)
        self.assertEquals(FooBar.renderStages, o.renderStages)

    def testRenderStageOverriding(self):
        o = mesh.newObject(FooBar).inRenderStages(
                IFancyVertexShader, IFancyFragmentShader)
        self.assertEquals((IFancyVertexShader, IFancyFragmentShader),
                o.renderStages)

    def testMaterialInclusion(self):

        o = mesh.newObject(FooBar)
        self.assertNot(o.material)

        o = mesh.newObject(FooBar).withMaterial(TestMaterial('bob', color=(0,0,0,1)))
        self.assert_(o.material)
        self.assertEquals(o.material.name, 'bob')
        self.assertEquals(o.material.color, (0,0,0,1))

        o = mesh.newObject(FooBar).withMaterial()
        self.assert_(o.material)


    def testGroupingAPI(self):
        o1 = mesh.newObject(FooBar)
        o1.pos = (1,1,1)
        o2 = mesh.newObject(FooBar)
        o2.pos = (1,2,3)
        o3 = mesh.newObject(FooBar)
        o3.pos = (4,5,6)
        group = mesh.Group(o1, o2, o3)
        group.pos += (2,3,4)

        # This really needs functional testing
        self.assertEquals((1,1,1), o1.pos)
        self.assertEquals((1,2,3), o2.pos)
        self.assertEquals((4,5,6), o3.pos)
        self.assertEquals((2,3,4), group.pos)


class BoundingBoxTest(unittest.TestCase):


    def _boundingBox(self, center, radius):
        return mesh.BoundingBox(
            (center.x-radius, center.x+radius),
            (center.y-radius, center.y+radius),
            (center.z-radius, center.z+radius))
        

    def testOverlapsContainsAndRadiusX(self):
        a = self._boundingBox(Vector3(0,0,0), 10)
        b = self._boundingBox(Vector3(20,0,0), 9)
        self.assertNot(a.overlaps(b) or b.overlaps(a))
        
        c = self._boundingBox(Vector3(20,0,0), 11)
        self.assert_(a.overlaps(c) and c.overlaps(a))
        self.assert_(c.overlaps(b) and b.overlaps(c))
        
        self.assert_(c.contains(b))
        self.assertNot(b.contains(c))
        
        d = self._boundingBox(Vector3(20,0,0), 10)
        self.assert_(a.overlaps(d) and d.overlaps(a))
        
        self.assert_(d.contains(b) and c.contains(d))
        self.assertNot(b.contains(d))
    
    def testCornerTouch(self):
        a = self._boundingBox(Vector3(0,0,0), 20)
        b = self._boundingBox(Vector3(40,40,40),20)
        b2 = self._boundingBox(Vector3(41,41,41),20)
        c = self._boundingBox(Vector3(20,20,20), 0.5)
        self.assert_(a.overlaps(b))
        self.assert_(b.overlaps(a))
        
        self.assertNot(a.overlaps(b2))
        self.assert_(c.overlaps(a) and a.overlaps(c))
        self.assert_(c.overlaps(b) and b.overlaps(c))
        self.assertNot(c.overlaps(b2) or b2.overlaps(c))
        
    def testObjectIntegration(self):
        o = mesh.newObject(FooBar)
        self.assertNot(o.boundingBox)

        o = mesh.newObject(FooBar)
        o.drawable.boundingBox = self._boundingBox(Vector3(0,0,0), 5)
        o.pos += (1,2,3)

        bb = o.boundingBox

        self.assertEquals(bb.minX, -4)
        self.assertEquals(bb.maxX, 6)
        self.assertEquals(bb.minY, -3)
        self.assertEquals(bb.maxY, 7)
        self.assertEquals(bb.minZ, -2)
        self.assertEquals(bb.maxZ, 8)
        
        
