"""Widgets for use in the environment's OSD.

Very simplistic/crude layout model currently:

* Pack algorithm per box with nested boxes
* Top-level pane which has 1 box
* Packing per box is Left to Right or Top to Bottom
"""

from pyglet.gl import *
from pyglet import font

from miru.components import implements
from miru.common import PositionalMixin

from miru import imiru

from miru.imiru import IDrawable, IDraggable, IWidget
from miru.imiru import ITranslationHandler, IRotationHandler
from miru.imiru import IOSDRenderStage
from miru import mesh
from miru import common

from miru.environment import env

from twisted.python.components import registerAdapter
from twisted.python import failure
from zope.interface import implementer

EXPAND = -1

ADD_OP = 0
REMOVE_OP = 1

DEFAULT_FONT = font.load('', 12)

class Box(object):
    implements(imiru.IHoverable, imiru.IPressable, imiru.IReleasable)

    padding = 5
    x = 0
    y = 0
    width = EXPAND
    height = EXPAND
    _last_over = None

    def __init__(self, **kw):
        self.objects = []
        self._hasboxes = None
        objects = kw.get('objects', [])
        for obj in objects:
            self.add(obj)
        self.padding = kw.get('padding', Box.padding)
        self._entered_descendants = set()

    def add(self, obj):
        if self._hasboxes is None:
            self._hasboxes = isinstance(obj, Box)
        else:
            assert isinstance(obj, Box) == self._hasboxes,\
                    'Boxes can contain generic widgets or boxes - not both'
        self.objects.append(obj)
    

    def repos(self, x, y):
        """Reposition the widget based at coordinates (x, y)
        """
        self.x = self.offx(x)
        self.y = self.offy(y)

    def offx(self, x):
        return x + (self.padding, 0)[self._hasboxes]

    def offy(self, y):
        raise NotImplementedError

    def pack(self, parent):
        raise NotImplementedError

    def draw(self, x, y):
        raise NotImplementedError

    def _check_entry(self, device, x, y, dx=0, dy=0, hover=False):
        w = None
        t = None
        for o in self.objects:
            try:
                w = imiru.IEnterable(o)
                if isinstance(o, Box):
                    wx = w.x
                    wy = w._punchy()
                else:
                    wx = w.x
                    wy = w.y
                if common.pointInRect(x, y, wx, wy,
                        w.width, w.height):
                    t = w
                    break
            except TypeError:
                pass
        if t:
            if t != self._last_over:
                t.on_device_enter(device, x, y)
                if self._last_over:
                    self._last_over.on_device_exit(device, x, y)
                    self.markEntered(self._last_over, REMOVE_OP)
                self._last_over = t
                self.markEntered(t, ADD_OP)
            try:
                w = imiru.IHoverable(t)
                w.on_device_hover(device, x, y, dx, dy)
            except TypeError:
                pass
        elif self._last_over:
            self._last_over.on_device_exit(device, x, y)
            self.markEntered(self._last_over, REMOVE_OP)
            self._last_over = None

    def on_device_hover(self, device, x, y, dx, dy):
        self._check_entry(device, x, y, dx=dx, dy=dy, hover=True)

    def on_device_enter(self, device, x, y):
        self._check_entry(device, x, y)

    def on_device_exit(self, device, x, y):
        self._check_entry(device, x, y)
        s = set(self._entered_descendants)
        for c in s:
            c.on_device_exit(device, x, y)
        s.clear()

    def on_device_press(self, device, x, y, button, modifiers):
        if not self._last_over: return
        try:
            w = imiru.IPressable(self._last_over)
            w.on_device_press(device, x, y, button, modifiers)
        except TypeError:
            pass

    def on_device_release(self, device, x, y, button, modifiers):
        if not self._last_over: return
        try:
            w = imiru.IReleasable(self._last_over)
            w.on_device_release(device, x, y, button, modifiers)
        except:
            pass

    def markEntered(self, descendant, op):
        if op == ADD_OP:
            self._entered_descendants.add(descendant)
        elif descendant in self._entered_descendants:
            self._entered_descendants.remove(descendant)
        self.parent.markEntered(descendant, op)

    # FIXME - remove this stupid _puncy hack make y the actual lower point
    # on the widget - draw will have to be modified to understand this better 
    # or something ...
    def _punchy(self):
        return self.y

    def _depth(self, depth=0):
        return self.parent._depth(depth+1)

class LRBox(Box):
    """Items are added left to right..
    """

    def pack(self, parent):
        self.parent = parent
        w = (self.padding, 0)[self._hasboxes]
        p = (self.padding, 0)[self._hasboxes]
        h = 0
        obj = None
        for obj in self.objects:
            if self._hasboxes:
                obj.pack(self)
            w += obj.width + p
            h = max(h, obj.height)
        self.width = w
        self.height = h + (p * 2)
        return self

    def offy(self, y):
        return y + (self.padding, 0)[self._hasboxes]

    def draw(self):
        offx = self.x
        paddx = (self.padding, 0)[self._hasboxes]
        offy = self.y
        for obj in self.objects:
            obj.repos(offx, offy)
            obj.draw()
            offx += paddx + obj.width


class TBBox(Box):
    """Items are added top to bottom.
    """

    def pack(self, parent):
        self.parent = parent
        h = (self.padding, 0)[self._hasboxes]
        p = (self.padding, 0)[self._hasboxes]
        w = 0
        obj = None
        for obj in self.objects:
            if self._hasboxes:
                obj.pack(self)
            h += obj.height + p
            w = max(w, obj.width)
        self.height = h
        self.width = w + (p * 2)
        return self
   
    def offy(self, y):
        #return y + (self.padding, 0)[self._hasboxes]
        return (y + self.height) - (self.padding, 0)[self._hasboxes]

    def _punchy(self):
        return self.y - self.height - self.padding

    def draw(self):
        offx = padx = self.x
        pady = (self.padding, 0)[self._hasboxes]
        offy = self.y
        for obj in self.objects:
            offy -= obj.height
            obj.repos(offx, offy)
            obj.draw()
            offy -= pady

class Filler:
    def __init__(self, width, height):
        self.width = width
        self.height = height
    def repos(self, x, y):
        pass
    def draw(self):
        pass

class ColorBar(object):
    """ColorBars are can be used to compose color adjustment widgets.
    All color 4-tuples are passed to glColor4f, so color should be 
    in range [0,1].

    @param top: Color 4-tuple for top_left and top_right
    @type top: tuple
    @param bottom: Color 4-tuple for bottom_left and bottom_right
    @type bottom: tuple
    @param left: Color 4-tuple for top_left and bottom_left
    @type left: tuple
    @param right: Color 4-tuple for top_right and bottom_right
    @type right: tuple
    @param top_left: Color 4-tuple for top_left
    @type top_left: tuple
    @param top_right: Color 4-tuple for top_right
    @type top_right: tuple
    @param bottom_left: Color 4-tuple for bottom_left
    @type bottom_left: tuple
    @param bottom_right: Color 4-tuple for bottom_right
    @type bottom_right: tuple

    >>> from miru.components import verifyClass, verifyObject
    >>> verifyClass(IDrawable, ColorBar)
    True
    >>> verifyObject(IDrawable, ColorBar(top=(1,1,1,1), bottom=(0,0,0,1)))
    True
    """
    implements(IDrawable, imiru.IEnterable, imiru.IExitable)
    x = 0
    y = 0
    width = 10
    height = 50

    def __init__(self, **kw):
        for k in ('top', 'bottom'):
            v = kw.get(k)
            if v is not None:
                setattr(self, '%s_left' % k, v)
                setattr(self, '%s_right' % k, v)
        for k in ('left', 'right'):
            v = kw.get(k)
            if v is not None:
                setattr(self, 'top_%s' % k, v)
                setattr(self, 'bottom_%s' % k, v)

        for k in ('top', 'bottom'):
            if not hasattr(self, '%s_left' % k):
                setattr(self, '%s_left' % k, kw.get('%s_left' % k))
            if not hasattr(self, '%s_right' % k):
                setattr(self, '%s_right' % k, kw.get('%s_right' % k))

        self.width = kw.get('width', self.width)
        self.height = kw.get('height', self.height)
        self.debug = kw.get('debug', False)

    def repos(self, x, y):
        self.x = x
        self.y = y

    def draw(self):
        x = self.x
        y = self.y
        glBegin(GL_QUADS)
        glColor4f(*self.bottom_left)
        glVertex3f(x, y, 0)
        glColor4f(*self.top_left)
        glVertex3f(x, y + self.height, 0)
        glColor4f(*self.top_right)
        glVertex3f(x + self.width, y + self.height, 0)
        glColor4f(*self.bottom_right)
        glVertex3f(x + self.width, y, 0)
        glEnd()


    def _c(self, channel):
        new_channel = (channel + 0.15) % 1
        return new_channel

    def on_device_enter(self, device, x, y):
        if self.debug:
            self._tl =  tl = self.top_left
            self._bl = bl = self.bottom_left
            self._tr = tr = self.top_right
            self._br = br = self.bottom_right

            c = self._c

            self.top_left = tuple([c(tl[i]) for i in range(4)])
            self.bottom_left = tuple([c(bl[i]) for i in range(4)])
            self.top_right = tuple([c(tr[i]) for i in range(4)])
            self.bottom_right = tuple([c(br[i]) for i in range(4)])

    def on_device_exit(self, device, x, y):
        if self.debug:
            self.top_left = self._tl
            self.bottom_left = self._bl
            self.top_right = self._tr
            self.bottom_right = self._br

class Button(ColorBar):
    implements(IWidget)
    width = 100
    height = 15
    top_left = top_right = (0.6,0.95,0.6,1)
    bottom_left = bottom_right = (0.1,0.17,0.1,1)
    top_left_hover = top_right_hover = (0.75,0.99,0.75,1)
    bottom_left_hover = bottom_right_hover = (0.05,0.07,0.05,1)
    top_left_down = top_right_down = (0.05,0.07,0.05,1)
    bottom_left_down = bottom_right_down = (0.75,0.99,0.75,1)
    _top_left = top_left
    _top_right = top_right
    _bottom_left = bottom_left
    _bottom_right = bottom_right

    def __init__(self, text, **kw):
        self.text = text
        self.display_text = font.Text(DEFAULT_FONT, text)
        self.callback = kw.get('callback', lambda btn : None)
        super(Button, self).__init__(**kw)

    def on_device_enter(self, device, x, y):
        self.top_left = self.top_left_hover
        self.top_right = self.top_right_hover
        self.bottom_left = self.bottom_left_hover
        self.bottom_right = self.bottom_right_hover

    def on_device_exit(self, device, x, y):
        #super(Button, self).on_device_exit(device, x, y)
        self.top_left = self._top_left
        self.top_right = self._top_right
        self.bottom_left = self._bottom_left
        self.bottom_right = self._bottom_right

    def on_device_release(self, device, x, y, button, modifiers):
        self.top_left = self.top_left_hover
        self.top_right = self.top_right_hover
        self.bottom_left = self.bottom_left_hover
        self.bottom_right = self.bottom_right_hover
        self.callback(self)

    def on_device_hover(self, device, x, y, dx, dy):
        pass

    def on_device_press(self, device, x, y, button, modifiers):
        self.top_left = self.top_left_down
        self.top_right = self.top_right_down
        self.bottom_left = self.bottom_left_down
        self.bottom_right = self.bottom_right_down

    def draw(self):
        super(Button, self).draw()
        t = self.display_text
        t.x = (self.x + self.width) - (self.width / 2.) - (t.width / 2.)
        t.y = self.y + 3
        t.draw()
        env.font_text = self.display_text

class Pane(ColorBar):
    """

    >>> from zope.interface.verify import verifyClass, verifyObject
    >>> verifyClass(IWidget, Pane)
    True
    >>> verifyObject(IWidget, Pane(box=None))
    True
    """
    implements(IDraggable, IWidget)
    renderStages = (IOSDRenderStage,)

    top_left = top_right = (0.8,0.8,0.95,0.9)
    bottom_left = bottom_right =  (0.15,0.05,0.15,0.6)
    width = 100
    height = 10
    _wrapper = None
    _last_over = None

    def __init__(self, **kw):
        if kw.has_key('padding'):
            self.padding = kw['padding']
        self.box = kw['box']
        self._entered_descendants = set()
        super(Pane, self).__init__(**kw)

    def pack(self):
        box = self.box
        box.pack(self)
        self.width = box.width
        self.height = box.height
        return self

    def draw(self):
        super(Pane, self).draw()
        self.x = self._wrapper.pos.x
        self.y = self._wrapper.pos.y
        self.box.repos(self.x, self.y)
        self.box.draw()

    def on_device_enter(self, device, x, y):
        self.top_left = self.top_right = (0.95,0.95,1,0.9)
        self.bottom_left = self.bottom_right = (0.3,0.1,0.3,0.7)
        device.focussed = self

    def on_device_exit(self, device, x, y):
        self.top_left = self.top_right = (0.8,0.8,0.95,0.9)
        self.bottom_left = self.bottom_right = (0.15,0.05,0.15,0.6)
        s = set(self._entered_descendants)
        for desc in s:
            desc.on_device_exit(device, x, y)
        self._entered_descendants.clear()

    def on_device_hover(self, device, x, y, dx, dy):
        self.box.on_device_hover(device, x, y, dx, dy)

    def on_device_press(self, device, x, y, button, modifiers):
        self.box.on_device_press(device, x, y, button, modifiers)

    def on_device_release(self, device, x, y, button, modifiers):
        self.box.on_device_release(device, x, y, button, modifiers)

    def markEntered(self, descendant, op):
        if op == ADD_OP:
            self._entered_descendants.add(descendant)
        elif descendant in self._entered_descendants:
            self._entered_descendants.remove(descendant)

    def _depth(self, depth=0):
        return depth


class Text:

    x = 0
    y = 0

    def __init__(self, text, width=100):
        self.text = ' '.join(text.splitlines())
        self.width = width
        self.height = 0 # dynamically determined
        self._lines = []
        self.reset()

    def reset(self):
        # FIXME - i18n support (`words' means something very different in CJK languages)
        words = self.text.split()
        line = []
        self.height = 0
        for word in words:
            l = list(line)
            line.append(word)
            display = font.Text(DEFAULT_FONT, ' '.join(line))
            if display.width > self.width:
                if not l:
                    l = [word] # Too bad if the word is too long for the area width
                    line = []
                else:
                    line = [word]
                display = font.Text(DEFAULT_FONT, ' '.join(l))
                self.height += display.height
                self._lines.append(display)
        if line:
            display = font.Text(DEFAULT_FONT, ' '.join(line))
            self.height += display.height
            self._lines.append(display)
        self._lines.reverse()


    def repos(self, x, y):
        self.x = x
        self.y = y

    def draw(self):
        x = self.x
        y = self.y
        for l in self._lines:
            l.x = x
            l.y = y
            l.draw()
            y += l.height


class TitleBar(ColorBar):
    implements(imiru.IWidget)
    bottom_left = bottom_right = (0.7,0.8,0.95,0.9)
    top_left = top_right =  (0.35,0.15,0.35,0.6)
    height=15

    def __init__(self, **kw):
        self.title = kw.get('title', u'')
        self.pane = kw['pane']
        self.title_display = font.Text(DEFAULT_FONT, self.title, color=(0,0,0,1))
        super(TitleBar, self).__init__(**kw)

    def draw(self):
        super(TitleBar, self).draw()
        self.title_display.x = self.x + 3
        self.title_display.y = self.y + 2
        self.title_display.draw()
    
    def on_device_press(self, device, x, y, button, modifiers):
        # The default behavior is just to make myself the target
        # of the handle
        o = self.pane._wrapper is None and self.pane or self.pane._wrapper
        device.target = o
        device.mode = device.TR_TRANSLATE
        # Move ourselves to the top of the OSD
        env.osd.remove(o)
        env.osd.append(o)
        common.changeCursor(env.window.CURSOR_HAND)


    def on_device_hover(self, device, x, y, dx, dy):
        pass

    def on_device_release(self, device, x, y, button, modifiers):
        pass

class MiruPane(Pane):
    """You can click it and drag it around and it has a little title
    bar with some text and minimize & close buttons.  Woo hoo!
    """
    def __init__(self, **kw):
        title = kw.get('title', u'')
        self.title_bar = TitleBar(title=title, pane=self)
        box = TBBox(objects=[
            LRBox(objects=[self.title_bar], padding=0),
            kw['box']], padding=0)
        kw['box'] = box
        super(MiruPane, self).__init__(**kw)

    def pack(self):
        super(MiruPane, self).pack()
        self.title_bar.width = self.box.objects[0].width = self.width
        self.box.objects[0].height = self.title_bar.height

# Widget builder utilities - API gloss ;)

def _noop(*p):
    pass


def _pane(klass):
    def f(box, **kw):
        kw['box'] = box
        o = mesh.newObject(klass, **kw).withTransformFunction(_noop)
        o.drawable.pack()
        return o
    return f

pane = _pane(Pane)

mpane = _pane(MiruPane)

def leftRight(*objects, **kw):
    kw['objects'] = objects
    return LRBox(**kw)


def topBottom(*objects, **kw):
    kw['objects'] = objects
    return TBBox(**kw)


@implementer(ITranslationHandler)
def _draggable_on_translate(handle, x, y, dx, dy):
    handle.target.pos += (dx, dy, 0)

def _draggable_on_translate_adpt(_context):
    return _draggable_on_translate

registerAdapter(_draggable_on_translate_adpt, IDraggable, ITranslationHandler)

@implementer(IRotationHandler)
def _draggable_on_rotate(handle, x, y, dx, dy): # a noop
    pass

def _draggable_on_rotate_adpt(_context):
    return _draggable_on_rotate

registerAdapter(_draggable_on_rotate_adpt, IDraggable, IRotationHandler)




