#from euclid import Vector3
import math
import euclid

from roboto import scene
from roboto import constants
clock = scene.scheduler

from twisted.internet import defer
from twisted.python import failure
from twisted.python import log

#from pyglet import clock

from zope.interface import Interface, Attribute
from zope.interface import implements

from miru.environment import env

FORWARD = 1
REVERSE = -1

class IActor(Interface):

    row = Attribute("""
            Current row coordinate of actor""")

    col = Attribute("""
            Current col coordinate of actor""")

    tile = Attribute("""
            Current tile actor is on""")

    path = Attribute("""Stack of positions.
        The first position, seconds and so on ...
        """)

    velocity = Attribute("""Current (positional) velocity
        vectoru (C{euclid.Vector3}) of the actor""")

    angle_velocity = Attribute("""Current angle velocity
        vector (C{euclid.Vector3}) of the actor""")

    def start(grid_coord):
        """This is called in C{importLevel} to position the 
        actor at the (row,col) coordinates given by C{grid_coord}.

        @param grid_coord The grid coordinates.
        """

    def act(action, index=-1):
        """Enact an action
        
        If index is -1, pause the window and start the action. When the action
        is complete, the action is pushed on the actions stack, the current
        (row,col) coordinates of the actor are pushed onto the path stack and
        the window is unpaused.

        If index is not -1, the window is first paused, the actor is positioned
        back at the position at location path[index], path is updated to 
        path[:index+1], the action stack is reset to actions[:index], 
        the actions popped from the stack and the current action
        (i.e., [action] + actions[index:]) are replayed 
        in order and when all replays have completed, the window is unpaused.


        Side Note: `pausing' the window is open to interpretation, but I think
        this should mean you can't go back and edit previous actions until the
        window is unpaused.  It would make sense to let the user continue to type
        new words, of course, but these should be buffered by a controller
        which will wait until outstanding actions are complete before calling act
        with the next backlogged action.  Where the said controller exists is
        also TBD - but it could just be the actor itself - where on calls to act
        it checks if it has an oustanding action and, just queues the action 
        in a FIFO and attaches a callback on the outstanding action which
        will just call act again.

        @param action: The action to perform

        @type action: C{actor.IActor}

        @param index: Index in action stack of the modified action.
        
        @return deferred that will fire on completion

        @rtype C{twisted.internet.defer.Deferred}
        """

    def rollback():
        """Rollback last action on the stack.  After completion, the last
        action on the actions stack is removed and the last position on the paths
        stakcs is removed.  Before rollback occurs, the window is paused.  After
        completion, the window is unpaused.

        @return deferred that will fire on completion

        @rtype C{twisted.internet.defer.Deferred}
        """

class IAction(Interface):

    actor = Attribute("""The C{IActor} associated with this action.
        """)
    
    # WARNING: direction isn't used yet ... may not really belong
    # in the interface.
    direction = Attribute("""Direction defined for the action.
        Possible values are: roboto.actor.FORWARD, roboto.actor.REVERSE
        """)

    def start():
        """Start the action and return a deferred that will fire
        with the updated actor object on completion.

        @return deferred that will fire on completion

        @rtype C{twisted.internet.defer.Deferred}
        """
    
    def stop():
        """Halt the action - generally unschedule from clock.  The action
        will try to stop itseld at conculsion.
        """

    def update(dt):
        """Update the action animation or whatever.

        @param dt: Time elapsed since last clock tick

        @type dt: C{float}
        """

class BaseAction(object):

    implements(IAction)

    direction = FORWARD
    actor = None
    running = False

    def start(self):
        self.running = True
        self.deferred = defer.Deferred()
        clock.schedule(self.update)
        return self.deferred

    def stop(self):
        self.running = False
        clock.unschedule(self.update)

    def fuckingStop(self):
        self.stop()
        if not self.deferred.called:
            self.deferred.callback(self)

    def update(self, dt):
        raise NotImplemented



class Teleport(BaseAction):

    def __init__(self, actor, row, col):
        self.actor = actor
        self.row = row
        self.col = col

    def update(self, dt):
        if not self.running:
            return
        try:
            (x,y,z) = scene.getPos(self.row, self.col)
            self.actor.actor_obj.pos = (x,y,z)
            self.stop()
            self.deferred.callback(self)
        except:
            self.stop()
            reason = failure.Failure()
            self.deferred.errback(reason)


class Actor(object):
    implements(IActor)

    in_action = False
    current_action = None

    row = 0
    col = 0
    tile = None
    current_deferred = None
    queued_callback = None

    def __init__(self, actor_obj):
      self.actor_obj = actor_obj

      self.velocity = euclid.Vector3(0,0,0)
      self.angle_velocity = euclid.Vector3(0,0,0)

      self.act_buffer = []
      self.actions = []
      #(x,y,z) = tile.getPos(0,0)
      self.path = [(0,0)] # Or whatever the start position is for the level ...


    def start(self, grid_coord=(0,0)):
        self.path = [grid_coord]
        self.row = grid_coord[0]
        self.col = grid_coord[1]
        self.actor_obj.pos = scene.getPos(*grid_coord)


    def act(self, action_spec, index=-1, deferred=None, insert=False):
        action = action_spec[0](self, *action_spec[1:])

        # FIXME buffer actions here if multiple added with index=-1
        #qd = self._queue(action, index)
        #if qd:
        #    return qd

        # I think we should pause the UI at this point to keep
        # the player from typing while the action is executing ... 
        # otherwise it's going to get messy to manage. (drew)
        env.window.pause()

        d = deferred or defer.Deferred()
        self.current_deferred = d

        def cb(action, unpause):
            self.actions.append(action_spec)
            self.path.append((self.row, self.col))
            if unpause:
                env.window.unpause()
                d.callback(self)
            self.current_action = None

        def eb(reason, index):
            # this needs some more thought
            print 'act fucked up', reason
            self.path = self.path[:index+1]
            self.actor_obj.pos = scene.getPos(*self.path[index])
            self.current_action = None
            # feed this back into the UI???

        if index > -1:
            offs = (index + 1, index)[insert]
            if not action:
                replay = self.actions[offs:]
            elif type(action) is list:
                replay = action + self.actions[offs:]
            else:
                replay = [action_spec] + self.actions[offs:]
            self.actions = self.actions[:index]
            self.path = self.path[:index+1]
            self.actor_obj.pos = scene.getPos(*self.path[-1])

            # This is an example of where inlineCallbacks would be nice :(
            def next(_action, remaining, idx, next_cb):
                if remaining:
                    next_spec = remaining.pop(0)
                    next_action = next_spec[0](self, *next_spec[1:])
                    #log.msg('replaying action: %r' % next)
                    self.current_action = next_action
                    next_action.start().addCallback(cb, False
                                ).addErrback(eb, idx
                                ).addCallback(next_cb, remaining, idx+1, next_cb
                                ).addErrback(eb, idx)
                else:
                    #d.callback(self)
                    self.current_deferred.callback(self)
                    env.window.unpause()
            next(None, replay, index, next)
            return d

        self.current_action = action
        action.start().addCallback(cb, True).addErrback(eb, len(self.path) - 1)
        return d


    def _queue(self, action, index):
        # this is where i'm starting to hurt on the queing stuff ... (drew)

        def drain_buffer(actor, ds):

            def allDone(actor):
                for d in ds:
                    d.callback(actor)

            minIdx = 999999
            for (action, idx) in self.act_buffer:
                minIdx = min(idx, minIdx)
                if action is None:
                    self.actions.pop(idx)
                else:
                    self.actions[idx] = action
            self.act_buffer = []
            d = defer.Deferred().addCallback(allDone)
            return self.act(self.actions[minIdx], index=minIdx, deferred=d)

        if self.current_deferred and not self.current_deferred.called:
            if not hasattr(self.current_deferred, 'drainActionDeferreds'):
                ds = []
                self.current_deferred.drainActionDeferreds = ds
                self.current_deferred.addBoth(drain_buffer, ds)
            ds = self.current_deferred.drainActionDeferreds
            ds.append(defer.Deferred())
            self.act_buffer.append((action, index))
            return ds[-1]


    def rollback(self):
        d = defer.Deferred()
        if self.actions:
            last = self.actions.pop()
            last_pos = self.path.pop()
            self.actor_obj.pos = scene.getPos(*self.path[-1])
            self.row = self.path[-1][0]
            self.col = self.path[-1][1]
        d.callback(self)
        return d



class ActionKiller(object):
    # Fucking stupid workaround

    _last = None
    count = constants.KILLER_COUNT

    def __init__(self):
        from pyglet import clock
        clock.schedule_interval(self.update, 1)


    def reset(self):
        self._last = None

    def update(self):
        
        if self._last is None:
            self._last = (env.window.actor.current_action, -1)
        action = env.window.actor.current_action
        if action is None:
            return
        lastAction, count = self._last
        print 'checking: %r - %d' % (lastAction, count)
        if lastAction == action:
            if count > self.count:
                print 'ERROR - killing unresponsive action', action
                action.stop()
                def doRollback(result):
                    print 'WARN - Rolling back actor', env.window.actor
                    env.window.actor.rollback()
                action.deferred.addBoth(doRollback)
                if not action.deferred.called:
                    action.deferred.callback(action)
            else:
                self._last = (action, count + 1)
        else:
            self._last = (action, -1)

actionKiller = ActionKiller()


###################################################################
#
# FIXME Improve Jump and Go to use velocity/acceleration 
#       Jump should use trajectories instead of stupid hack
#
###################################################################

class Jump(BaseAction):
    moving_up = True
  
    def __init__(self, actor, length, height):
        self.actor = a = actor
        o = self.actor.actor_obj
        self.length = length
        self.height = height
        self.tgtrow, self.tgtcol = self._getTarget()
        self.tgtpos = euclid.Vector3(*scene.getPos(self.tgtrow, self.tgtcol))
        self.midx = (o.pos.x + self.tgtpos.x) / 2.0
        self.midz = (o.pos.z + self.tgtpos.z) / 2.0

    def update(self, dt):
        o = self.actor.actor_obj
        if (self.tgtrow - self.actor.row) > 0:
            o.pos += (0.05, 0, 0)
            if o.pos.x < self.midx:
                o.pos += (0, 0.1, 0)
            else:
                o.pos -= (0, 0.1, 0)
        elif (self.tgtrow - self.actor.row) < 0:
            o.pos -= (0.05, 0, 0)
            if o.pos.x > self.midx:
                o.pos += (0, 0.1, 0)
            else:
                o.pos -= (0, 0.1, 0)
        if (self.tgtcol - self.actor.col) < 0:
            o.pos += (0, 0, 0.05)
            if o.pos.z < self.midz:
                o.pos += (0, 0.1, 0)
            else:
                o.pos -= (0, 0.1, 0)
        elif (self.tgtcol - self.actor.col) > 0:
            o.pos -= (0, 0, 0.05)
            if o.pos.z > self.midz:
                o.pos += (0, 0.1, 0)
            else:
                o.pos -= (0, 0.1, 0)
        if self.atTarget():
            x = self.tgtpos.x
            z = self.tgtpos.z
            y = self.tgtpos.y
            o.pos = (x, y, z)
            a = self.actor
            a.tile = scene.getTile(self.tgtrow, self.tgtcol)
            a.row = self.tgtrow
            a.col = self.tgtcol
            self.stop()
            self.deferred.callback(self)


    def _getTarget(self):
        a = self.actor
        ang = a.actor_obj.angle.y
        length = self.length
        rows, cols = scene.getFloorDimensions()
        tgt = (a.row, a.col)
        if ang == 0:
            if (a.row + length) < rows - 1:
                tgt = (a.row + length, a.col)
            else:
                tgt = (rows - 1, a.col)
        elif ang == 90:
            if (a.col + length) < cols - 1:
                tgt = (a.row, a.col + length)
            else:
                tgt = (a.row, cols - 1)
        elif ang == 180:
            if (a.row - length) > 0:
                tgt = (a.row - length, a.col)
            else:
                tgt = (0, a.col)
        elif ang == 270:
            if (a.col - length) > 0:
                tgt = (a.row, a.col - length)
            else:
                tgt = (a.row, 0)
        return tgt 

    def atTarget(self):
        o = self.actor.actor_obj
        if (o.pos.x <= self.tgtpos.x + 0.1 and o.pos.x >= self.tgtpos.x - 0.1) \
           and (o.pos.z <= self.tgtpos.z + 0.1 and o.pos.z >= self.tgtpos.z - 0.1):
            return True


class Go(BaseAction):
    
    def __init__(self, actor, length):
        self.actor = a = actor
        self.length = length
        self.tgtrow, self.tgtcol = self._getTarget()
        self.tgtpos = euclid.Vector3(*scene.getPos(self.tgtrow, self.tgtcol))

    def update(self, dt):
       o = self.actor.actor_obj
       if (self.tgtrow - self.actor.row) > 0:
           o.pos += (0.05, 0, 0)
       elif (self.tgtrow - self.actor.row) < 0:
           o.pos -= (0.05, 0, 0)
       if (self.tgtcol - self.actor.col) < 0:
           o.pos += (0, 0, 0.05)
       elif (self.tgtcol - self.actor.col) > 0:
           o.pos -= (0, 0, 0.05)
       if self.atTarget():
           x = self.tgtpos.x
           z = self.tgtpos.z
           o.pos = (x, o.pos.y, z)
           a = self.actor
           a.tile = scene.getTile(self.tgtrow, self.tgtcol)
           a.row = self.tgtrow
           a.col = self.tgtcol
           self.stop()
           self.deferred.callback(self)

    def atTarget(self):
       o = self.actor.actor_obj
       if (o.pos.x <= self.tgtpos.x + 0.1 and o.pos.x >= self.tgtpos.x - 0.1) \
          and (o.pos.z <= self.tgtpos.z + 0.1 and o.pos.z >= self.tgtpos.z - 0.1):
           return True
           
    def _getTarget(self):
       a = self.actor
       ang = a.actor_obj.angle.y
       length = self.length
       rows, cols = scene.getFloorDimensions()
       tgt = (a.row, a.col)
       if ang == 0:
           if (a.row + length) < rows - 1:
               tgt = (a.row + length, a.col)
           else:
               tgt = (rows - 1, a.col)
       elif ang == 90:
           if (a.col + length) < cols - 1:
               tgt = (a.row, a.col + length)
           else:
               tgt = (a.row, cols - 1)
       elif ang == 180:
           if (a.row - length) > 0:
               tgt = (a.row - length, a.col)
           else:
               tgt = (0, a.col)
       elif ang == 270:
           if (a.col - length) > 0: 
               tgt = (a.row, a.col - length)
           else:
               tgt = (a.row, 0)
       return tgt

class Spin(BaseAction):

    def __init__(self, actor, numtimes):
        self.actor = actor
        self.degrees = numtimes*(360.0)

    def update(self, dt):
        self.actor.actor_obj.angle.y += 6.0
        if self.actor.actor_obj.angle.y > self.degrees:
            self.actor.actor_obj.angle.y = self.degrees
            self.actor.actor_obj.angle.y %= 360 # keep actor angle in 0-360
            self.stop()
            self.deferred.callback(self)

class Turn(BaseAction):
    
    def __init__(self, actor, degrees):
        self.actor = actor
        self.degrees = degrees
        self.old = self.actor.actor_obj.angle.y

    def update(self, dt):
        o = self.actor.actor_obj
        if self.degrees >= 0:
            o.angle.y += 1.0
        elif self.degrees < 0:
            o.angle.y -= 1.0
        if self.degrees >= 0:
            if o.angle.y > (self.degrees + self.old):
                o.angle.y = self.degrees + self.old
                o.angle.y %= 360
                self.stop()
                self.deferred.callback(self)
        elif self.degrees < 0:
            if o.angle.y < (self.degrees + self.old):
                o.angle.y = self.degrees + self.old
                o.angle.y %= 360
                self.stop()
                self.deferred.callback(self)


