import pyglet.window.key as key
from pyglet.font import Text

from miru.environment import env

from roboto import vocable,events
from roboto.constants import *
from roboto.levels import menulevel

class NonsenseEditor(object):
    
    def __init__(self, text=''):
        self.loadFonts()
        # add initial empty word
        self.focus = None
        self.shown = False
        self.setText(text)
        events.levelTransitionEvent.push_handlers(on_transition=self.on_level_transition)
    
    def loadFonts(self):
        from pyglet.font import load as load_font
        self.smallFont = load_font(NONSENSE_EDITOR_FONT, SMALL_NONSENSE_TEXT_SIZE)
        self.largeFont = load_font(NONSENSE_EDITOR_FONT, LARGE_NONSENSE_TEXT_SIZE)
        
        self.cursor = Text(self.smallFont, '^')
    
    def on_level_transition(self, _1, _2):
        self.setText('')
        if env.levels.current != menulevel:
            self.show()
        else:
            self.hide()
    
    #
    # External interface
    #
    def setText(self, text):
        words = text.split()
        if not words:
            words = ['']
        
        self.words = [NonsenseWord(word,self) for word in words]
        self.currentWordIndex = len(self.words)-1
        self.currentWord().focusOnRight()
        self.recalcTextPosition()
    
    def show(self):
        "Attach event listeners and display on the screen"
        if not self.shown:
            env.osd.addobj(self)
            env.window.push_handlers(self.on_text, self.on_text_motion)
            self.shown = True
        
    def hide(self):
        "Remove event listeners and hide"
        if self.shown:
            env.osd.delobj(self)
            #env.window.pop_handlers() # TODO: remove by name
            self.shown = False
        
    
    #
    # Visual properties
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
    def draw(self):
        currentX = self.textStartX
        for word in self.words:
            currentX = word.draw(currentX)
        
        self.cursor.x = self.textStartX + self._widthToCursor + CURSOR_X_OFFSET
        self.cursor.y = CURSOR_Y_OFFSET
        self.cursor.draw()
        
    def width(self, words=None):
        if words is None:
            words = self.words
        wordWidths = sum([word.width() for word in words])
        spaceWidths = NONSENSE_SPACE_WIDTH*(len(words)-1)
        return wordWidths + spaceWidths
    
    def widthToCursor(self):
        prevWordsWidth = self.width(self.words[:self.currentWordIndex])
        return prevWordsWidth + self.currentWord().widthToCursor()
    
    def recalcTextPosition(self):
        width = self.widthToCursor()
        if env.window:
            # hmm!
            visibleWidth = env.window.width-HUD_SIDES_PADDING*2
        else:
            visibleWidth = 2000
        
        if width > visibleWidth:
            self.textStartX = env.window.width-width-HUD_SIDES_PADDING
        else:
            self.textStartX = HUD_SIDES_PADDING
        self._widthToCursor = width
    
    #
    # Input logic
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
    def on_text(self, text):
        initialWordIndex = self.currentWordIndex
        currentWord = self.currentWord()
        if text == ' ':
            
            if currentWord.text:
                if currentWord.atRight() and self.rightWordIsEmpty():
                    self.currentWordIndex += 1
                    self.currentWord().focus(0)
                    # moving rightward to an empty space
                                
                elif currentWord.atLeft():
                    if not self.leftWordIsEmpty():
                        self.words.insert(self.currentWordIndex, NonsenseWord('',self))
                    else:
                        self.currentWordIndex -= 1
                    self.currentWord().focus(0)
                    # moving leftward to a (possibly new) empty space
                
                else:
                    tail = currentWord.split()
                    self.words.insert(self.currentWordIndex+1, tail)
                    self.currentWordIndex += 1
                    self.currentWord().focusOnLeft()
                    # just spliting a word (or adding a word at the far right)
            
            self.afterWordMotion(initialWordIndex)
        
        else:
            text = text.lower()
            if text in ALLOWED_NONSENSE_LETTERS:
                self.words[self.currentWordIndex].addText(text)
        
        self.recalcTextPosition()
    
    
    def on_text_motion(self, motion):
        initialWordIndex = self.currentWordIndex
        currentWord = self.currentWord()
        
        if motion == key.MOTION_LEFT:
            if not currentWord.goLeft():
                if not self.atLeftmostWord():
                    self.currentWordIndex -= 1
                    self.currentWord().focusOnRight()
                    
        elif motion == key.MOTION_RIGHT:
            if not currentWord.goRight():
                if not self.atRightmostWord():
                    self.currentWordIndex += 1
                    self.currentWord().focusOnLeft()
                
        elif motion == key.MOTION_PREVIOUS_WORD:
            if self.atLeftmostWord():
                currentWord.focusOnLeft()
            else:
                if currentWord.atLeft():
                    self.currentWordIndex -= 1
                self.currentWord().focusOnLeft()

        elif motion == key.MOTION_NEXT_WORD:
            if self.atRightmostWord():
                currentWord.focusOnRight()
            else:
                if currentWord.atRight():
                    self.currentWordIndex += 1
                self.currentWord().focusOnRight()

        elif motion == key.MOTION_BEGINNING_OF_LINE:
            self.currentWordIndex = 0
            self.currentWord().focusOnLeft()
            
        elif motion == key.MOTION_END_OF_LINE:
            self.currentWordIndex = len(self.words)-1
            self.currentWord().focusOnRight()

        elif motion == key.MOTION_BACKSPACE:
            if currentWord.atLeft():
                if not self.atLeftmostWord():
                    self.currentWordIndex -= 1
                    leftWord = self.currentWord()
                    leftWord.focusOnRight()
                    leftWord.merge(currentWord)
                    self.words.remove(currentWord)
            else:
                currentWord.backspace()
        
        elif motion == key.MOTION_DELETE:
            if currentWord.atRight():
                if not self.atRightmostWord():
                    rightWord = self.words[self.currentWordIndex+1]
                    currentWord.merge(rightWord)
                    self.words.remove(rightWord)
            else:
                currentWord.delete()
        
        self.recalcTextPosition()
        if initialWordIndex != self.currentWordIndex:
            self.afterWordMotion(initialWordIndex)
    
    def afterWordMotion(self, initialWordIndex, insert=False):
        curWord = self.words[self.currentWordIndex]
        curAction = vocable.resolveOnomatopoeia(curWord.text, env.window.actor)
        
        if initialWordIndex < len(self.words):
            prevWord = self.words[initialWordIndex]
            prevAction = vocable.resolveOnomatopoeia(prevWord.text, env.window.actor)
            # we should put this in the queue
            env.window.actor.act(prevAction, initialWordIndex)
        # this is the one we should goto
        #env.window.actor.act(curAction, self.currentWordIndex)
    
    #
    # Input helpers
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
    def currentWord(self):
        return self.words[self.currentWordIndex]
    
    def leftWordIsEmpty(self):
        try:
            return self.words[self.currentWordIndex-1].text == ''
        except:
            return False
    
    def rightWordIsEmpty(self):
        try:
            return self.words[self.currentWordIndex+1].text == ''
        except:
            return False
    
    def atLeftmostWord(self):
        return self.currentWordIndex == 0
    
    def atRightmostWord(self):
        return self.currentWordIndex+1 == len(self.words)    




class NonsenseWord(object):
    "One word, that can be edited and displayed in the NonsenseEditor"
    def __init__(self, text, parent):
        self.currentIndex = 0
        self.text = text
        self.parent = parent
        self.hasFocus = False
        self.remakeLabel()
    
    #
    # Visual properties
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
    def remakeLabel(self):        
        self.label = Text(self.font(), self.text, y=NONSENSE_TEXT_BASELINE)

    def font(self):
        if self.hasFocus:
            return self.parent.largeFont
        else:
            return self.parent.smallFont
    
    def draw(self, xPos):
        self.label.x = xPos
        self.label.draw()        
        return xPos + self.label.width + NONSENSE_SPACE_WIDTH
    
    def width(self):
        return self.label.width
    
    def widthToCursor(self):
        return Text(self.font(), self.text[:self.currentIndex]).width
    
    #
    # Input/editing logic
    # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
    def addText(self, text):
        cur = self.currentIndex
        self.text = self.text[:cur] + text + self.text[cur:]
        self.currentIndex += 1
        self.remakeLabel()
    
    def delete(self):
        self.text = self.text[:self.currentIndex]+self.text[self.currentIndex+1:]
        self.remakeLabel()
    
    def backspace(self):
        self.text = self.text[:self.currentIndex-1]+self.text[self.currentIndex:]
        self.currentIndex -= 1
        self.remakeLabel()
    
    def split(self):
        child = NonsenseWord(self.text[self.currentIndex:], self.parent)
        self.text = self.text[:self.currentIndex]
        self.remakeLabel()
        return child
    
    def merge(self, other):
        self.text = self.text+other.text
        self.remakeLabel()
    
    def goLeft(self):
        if self.currentIndex > 0:
            self.currentIndex -= 1
            return True
        else:
            return False
    
    def goRight(self):
        if self.currentIndex < len(self.text):
            self.currentIndex += 1
            return True
        else:
            return False
    
    def atLeft(self):
        return self.currentIndex == 0
    
    def atRight(self):
        return self.currentIndex == len(self.text)
    
    def focus(self, index):
        focus = self.parent.focus
        if focus:
            focus.hasFocus = False
            focus.remakeLabel()
        
        self.parent.focus = self
        self.hasFocus = True
        self.currentIndex = index
        self.remakeLabel()
    
    def focusOnLeft(self):
        self.focus(0)
    
    def focusOnRight(self):
        self.focus(len(self.text))

nonsenseEditor = NonsenseEditor()
 
