"""
N.B.  The grid layout is composed of rows that run front to back
and columns that run left to right.  So, to access the front-most
and left-most tile:

    getTile(0,0)

To access the tile behind (away from the player) it:

    getTile(0,1)

And to access the back-most and right most tile (for a grid of
dimensions: NxM)

   getTile(N-1,M-1)
"""

from miru import mesh
from miru import imiru
from miru import camera
from miru import track

from roboto import constants
import roboto.tile

from zope.interface import implements

from pyglet import clock

import itertools

import os
P = os.path.join


TILE_WIDTH = roboto.tile.TILE_WIDTH
TILE_LENGTH = roboto.tile.TILE_LENGTH
TILE_SPACING = roboto.tile.TILE_SPACING

_loaders = {}
_importers = {}
_destroyers = {}
_lighgroup_loaders = {}
_light_loaders = {}
_actors = []

_levelImportCallbacks = []

_floor = None
_loaded_tgroups = {}


def getXZ(row, col):
    """Given (row,col) coordinates in matrix, return
    XY position in world coordinates
    """
    from warnings import warn
    warn("Don't use getXZ - use getPos instead")
    x = row * TILE_LENGTH
    if row > 1:
        x += (row-1) * (TILE_SPACING)
    y = col * TILE_WIDTH
    if col > 1:
        y += (col-1) * (TILE_SPACING)
    return (x,y)


def getPos(row, col):
    x = (row * TILE_LENGTH)
    z = -(col * TILE_WIDTH)
    return (x,0,z)

def tileGroup(grid, **kw):
    """Create tile group based on rows

    @param rows: 2D row-column matrix containing tile keys
    """
    g = [ [load(tile, rowi, tilei) for (tilei, tile) in enumerate(row)]  \
            for (rowi, row) in enumerate(grid) ]
    obj = mesh.newObject(roboto.tile.TileGroup, g, **kw)
    for o in obj.drawable._objects:
        o.group = obj.drawable
    return obj


def getFloor():
    """Get the current floor

    @return The floor (C{Object} wrapping C{TileGroup})
    @rtype C{miru.mesh.Object} 
    """
    return _floor

def getFloorDimensions():
    """Get (row count, column count) for the current floor
    """
    return (len(_floor.grid), len(_floor.grid[0]))

def isValidLocation(row,col):
  return row >= 0 and col >= 0 and row < len(_floor.grid) and col < len(_floor.grid[0])

def getTile(row, col):
    """Get the tile at (row,col).

    @param row: The row coordinate

    @param col: The column coordinate

    @return The tile at row,col
    @rtpye C{miru.mesh.Object}
    """
    return _floor.grid[row][col]

def registerLoader(key, loader):
    """Register a custom loader function for loading tiles with key.

    @param key: key of tile to associate custom loader with

    @param loader: loader function taking one argument (key)
    """
    _loaders[key] = loader

def _defaultLoad(key, row, col):
    objf = 'tile-%s.obj' % key
    obj = mesh.loadObj(P('data', objf))
    #group = getFloor().drawable
    obj.tile = roboto.tile.Tile(obj, row, col)
    return obj

def load(key, row, col):
    """Load an individual tile for the given key

    @param key: The tile to load

    @type key: C{str}
    """
    loader = _loaders.get(key, _defaultLoad)
    return loader(key, row, col)


def registerImporter(key, importer):
    """Register an level importer for key. The key should be assigned to
    the constant LEVEL_TYPE in the module to import.

    @param key: the key associated with the importer.
                (LEVEL_TYPE value in the level module)

    @type key: C{str}

    @param importer: importer function that takes two arguments:
        module and environment
    """
    _importers[key] = importer


def registerDestroyer(key, destroyer):
    """Register a level destroyer for key.  The key should be assigned to
    the constant LEVEL_TYPE in the module to destroy.

    @param key: The key associated with the level to destroy.

    @type key: C{str}

    @param destroyer: destroyer function that takes two arguments:
        module (C{module}) and environment (C{miru.environment.Environment})
    """
    _destroyers[key] = destroyer

def importLevel(mod, env):
    """Import level defined in mod into environment

    @param mod: module object containing level definitions

    @param env: C{Environment} instance
    """
    global _levelImportCallbacks
    ltype = getattr(mod, 'LEVEL_TYPE', 'level')
    import_ = _importers.get(ltype, _levelImporter)
    import_(mod, env)
    for (f,a,kw) in _levelImportCallbacks:
        f(*a, **kw)
    _levelImportCallbacks = []

def _levelImporter(mod, env):
    global _floor
    tgroup = tileGroup(mod.TILES, loc='floor')
    env.addobj(tgroup)
    _loaded_tgroups[id(mod.TILES)] = tgroup
    _floor = tgroup.drawable

    wall = tileGroup(mod.BACKWALL_TILES, loc='back')
    objs = tgroup.drawable._objects
    r = len(tgroup.drawable.rows[0])
    offz = TILE_WIDTH * r + ((r-1) * TILE_SPACING)
    wall.pos += (0,TILE_LENGTH/2.,(TILE_WIDTH/2.)-offz)
    wall.angle += (90,0,0)
    env.addobj(wall)
    _loaded_tgroups[id(mod.BACKWALL_TILES)] = wall


    wall = tileGroup(mod.LEFTWALL_TILES, loc='left')
    r = len(tgroup.drawable.rows)
    wall.angle += (90,90,0)
    wall.pos -= (TILE_LENGTH/2., -TILE_WIDTH/2., 0)
    env.addobj(wall)
    _loaded_tgroups[id(mod.LEFTWALL_TILES)] = wall
    
    
    wall = tileGroup(mod.RIGHTWALL_TILES, loc='right')
    offx = TILE_LENGTH * (r-1) + (r-1) * TILE_SPACING
    wall.angle += (90,90,0)
    wall.pos += (offx + TILE_LENGTH/2., TILE_WIDTH/2., 0)
    env.addobj(wall)
    _loaded_tgroups[id(mod.RIGHTWALL_TILES)] = wall

    from roboto import actor
    actor.actionKiller.reset()
    env.window.actor.tile = None


    # Set up the lighting
    light_key = getattr(mod, 'LIGHTS', 'default')
    if type(light_key) is str:
        loadLights = _lighgroup_loaders.get(light_key, _defaultLightGroupLoader)
        loadLights(mod, env)
    else:
        env.camera.lights = camera.LightGroup([])
        print 'Light loaders', _light_loaders
        for key in light_key:
            loadLight = _light_loaders.get(key)
            if loadLight:
                loadLight(mod, env)
            else:
                print 'ERROR! No light loader registed for %s' % key

    scheduler.activate()
    roboto.tile.tileEventDistpatcher.start()
    print 'scheduler activated'
    env.window.unpause()

registerImporter('level', _levelImporter)


def _menuImporter(mod, env):
    from roboto import menu
    env.menu = menu.Menu((200,200))
    _levelImporter(mod, env)
    env.menu.show()
  
registerImporter('menu', _menuImporter)


def registerLevelImportCallback(f, *a, **kw):
    _levelImportCallbacks.append((f, a, kw))

def destroyLevel(mod, env):
    """Destroy level objects previously loaded from module.

    @param mod: module object containing level definitions

    @param enf: C{Environment} instance
    """
    ltype = getattr(mod, 'LEVEL_TYPE', 'level')
    destroy = _destroyers.get(ltype, _levelDestroyer)
    destroy(mod, env)

def _levelDestroyer(mod, env):
    global _actors
    print '-' * 50

    for t in (mod.TILES, mod.BACKWALL_TILES,
            mod.LEFTWALL_TILES, mod.RIGHTWALL_TILES):
        tgroup = _loaded_tgroups.get(id(t))
        if tgroup:
            env.delobj(tgroup)
            del _loaded_tgroups[id(t)]
    for mob in _actors:
        mob.stop()
        scheduler.unschedule(mob.update)
        while mob in env.window.actors:
            env.window.actors.remove(mob)
        env.delobj(mob.actor.actor_obj)
    _actors = []

    for l in env.camera.lights:
        l.debug = False

    scheduler.deactivate()
    print 'Unschedule called'

registerDestroyer('level', _levelDestroyer)


def _menuDestroyer(mod, env):
    env.menu.hide()  
    print "Menu destroyed!"

registerDestroyer('menu', _menuDestroyer)


def registerLightLoader(key, loader):
    print 'Registering (%s, %r)' % (key, loader)
    _light_loaders[key] = loader

def registerLightGroupLoader(key, loader):
    _lighgroup_loaders[key] = loader


_hoverLight = camera.PositionalLight(diffuse=(0.1,0.9,0.2,1))
def _attachRoboHoverLight(mod, env):
    global _hoverLight
    robobj = env.window.actor.actor_obj
    #spotLight = camera.PositionalLight(diffuse=(0.1,0.9,0.2,1), debug=constants.DEBUG)
    spotLight = _hoverLight
    track.deactivate(tracked=spotLight)
    _hoverLight = camera.PositionalLight(diffuse=(0.1,0.9,0.2,1))
    spoltLight = _hoverLight
    track.PosTrack(spotLight, robobj)
    spotLight.pos = robobj.pos
    spotLight.pos += (0,0.5,0)
    robobj.pos -= (0,0.5,0)
    spotLight.pos += (0,1.1,0)
    env.camera.lights.append(spotLight)


registerLightLoader('robo-hover-light', _attachRoboHoverLight)

def _defaultLightGroupLoader(mod, env):
    env.camera.lights = camera.LightGroup([
        camera.PositionalLight(pos=(-0.29, 0.76, -1.73), degug=constants.DEBUG),
        camera.PositionalLight(pos=(5,     1.1,  -3.46), debug=constants.DEBUG),
        camera.DirectionalLight(ambient=(0.2,0.2,0.9,1), diffuse=(0.2,0.2,0.9,1)),
    ])
    _attachRoboHoverLight(mod, env)

registerLightGroupLoader('default', _defaultLightGroupLoader)

# A few extra:


def currentTileGroups():
    return list(_loaded_tgroups.values())


class RobotoScheduler(object):
    """Use me to schedule/unschedule events instead of pyglet.clock.  I will of course
    just use pyglet.clock for the actual scheduling, but keep track of references
    so I can clean up outstanding events at the end of the level without interfering
    with global events.
    """

    active = False

    def __init__(self):
        self._scheduled = set()

    def activate(self):
        self.active = True

    def deactivate(self):
        for f in self._scheduled:
            clock.unschedule(f)
        self.active = False
        self._scheduled = set()

    def schedule(self, f, *a, **kw):
        if not self.active:
            return
        clock.schedule(f, *a, **kw)
        self._scheduled.add(f)

    def unschedule(self, f):
        if not self.active:
            return
        clock.unschedule(f)
        if f in self._scheduled:
            self._scheduled.remove(f)

    def schedule_interval(self, f, interval, *a, **kw):
        if not self.active:
            return
        clock.schedule_interval(f, interval, *a, **kw)
        self._scheduled.add(f)

    def schedule_once(self, f, delay, *a, **kw):
        if not self.active:
            return
        clock.schedule_once(f, delay, *a, **kw)
        self._scheduled.add(f)

    
scheduler = RobotoScheduler()



def getRowCol(pos, actor=None):

    rows, cols = getFloorDimensions()
    x, _y, z = tuple(pos)

    row = int(x / TILE_LENGTH)
    col = int(abs(z) / TILE_WIDTH)

    if actor:
        bump = False
        if row >= rows:
            row = rows - 1
            bump = True
        if row < 0:
            row = 0
            bump = True
        if col >= cols:
            col = cols - 1
            bump = True
        if col < 0:
            col = 0
            bump = True

        if bump:
            actor.pos = getPos(row, col)
            actor.row = row
            actor.col = col
    return row, col



######################
# Level Generator    #
######################

class LevelGenerator(object):

    current = None

    def __init__(self, levels=None):
        if not levels:
            from roboto.levels import menulevel, level1, level2, level3, level4, level5
            levels = [menulevel, level2, level1, level2, level3, level4, level5]
        self._levels = itertools.cycle(levels)

    def next(self):
        self.current = self._levels.next()
        return self.current

from miru.environment import env
# Setup the default
env.levels = LevelGenerator()


