from twisted.trial import unittest
from twisted.internet import defer
from twisted.python import failure
from twisted.python import log

from miru.environment import env
from miru import mesh

from pyglet.window import Window

import roboto.actor
from roboto import scene

import os
P = os.path.join

_copied = False

class MockWindow(Window):

    paused = False

    def pause(self):
        self.paused = True

    def unpause(self):
        self.paused = False

mockWindow = MockWindow()


class HappyAction(roboto.actor.BaseAction):
    # Look suspiciously like Transport?

    def __init__(self, actor, row, col):
        self.actor = actor
        self.row = row
        self.col = col


    def start(self):
        self.deferred = defer.Deferred()
        return self.deferred

    def stop(self):
        pass

    def update(self, dt):
        try:
            tgt = scene.getTile(self.row, self.col)
            (x,y,z) = scene.getPos(self.row, self.col)
            self.actor.actor_obj.pos = (x,y,z)
            self.actor.tile = tgt
            self.actor.row = self.row
            self.actor.col = self.col
            self.stop()
            self.deferred.callback(self)
        except:
            self.stop()
            reason = failure.Failure()
            self.deferred.errback(reason)


class ActorTest(unittest.TestCase):


    def setUp(self):
        # XXX some dirty setup 
        global _copied
        if not _copied:
            import shutil
            shutil.copytree(P('..', 'data'), 'data')
        if os.path.exists('data'):
            _copied = True

        self._w = env.window
        env.window = mockWindow
        self._tw = scene.TILE_WIDTH
        self._tl = scene.TILE_LENGTH
        scene.TILE_WIDTH = 0.918224
        scene.TILE_LENGTH = 0.918224
        # Import the test level
        import testlevel
        scene.importLevel(testlevel, env)

    def tearDown(self):
        scene.TILE_WIDTH = self._tw
        scene.TILE_LENGTH = self._tl
        env.window = self._w
        import testlevel
        scene.destroyLevel(testlevel, env)

    def testInitialization(self):
        actor = roboto.actor.Actor(mesh.loadObj(P('data', 'lilbot.obj')))
        actor.start((1,2))
        self.assertEquals([(1,2)], actor.path)
        self.assertEquals([], actor.actions)

    def testActSingle(self):

        actor = roboto.actor.Actor(mesh.loadObj(P('data', 'lilbot.obj')))
        actor.start((0,0))

        action = HappyAction(actor, 1, 3)

        def cb(the_actor):
            self.assertEquals(id(the_actor), id(actor))
            self.assertEquals([(0,0),(1,3)], actor.path)
            self.assertEquals([action], actor.actions)

        def eb(reason):
            self.fail("%r" % reason)

        d = actor.act(action).addCallbacks(cb, eb)
        # Fake the event
        action.update(1.04234)
        return d



    def testActMuliaction(self):

        actor = roboto.actor.Actor(mesh.loadObj(P('data', 'lilbot.obj')))
        actor.start((0,0))

        action1 = HappyAction(actor, 4, 4)
        action2 = HappyAction(actor, 1, 3)

        def cb(the_actor, expected_path, expected_actions, cb_):
            self.assertEquals(id(the_actor), id(actor))
            self.assertEquals([(0,0)] + expected_path, actor.path)
            self.assertEquals(expected_actions, actor.actions)
            if len(expected_actions) == 1:
                d = actor.act(action2).addCallback(
                        cb_, [(4,4),(1,3)], [action1, action2], None
                        ).addErrback(eb)
                action2.update(0.043216546)
                return d

        def eb(reason):
            reason.printTraceback()
            self.fail("Fuck me")

        d = actor.act(action1).addCallback(
                cb, [(4,4)], [action1], cb
                ).addErrback(eb)
        # Fake the event
        action1.update(0.04327647)
        return d


    def testEditPath(self):

        actor = roboto.actor.Actor(mesh.loadObj(P('data', 'lilbot.obj')))
        actor.start((0,0))
      
        
        action1 = HappyAction(actor, 4, 4)
        actor.act(action1)
        action1.update(0.04327647)

        action2 = HappyAction(actor, 1, 3)
        actor.act(action2)
        action2.update(0.03426756)

        action3 = HappyAction(actor, 2, 0)
        actor.act(action3)
        action3.update(0.03426432)

        def cb(the_actor):
            self.assertEquals(id(the_actor), id(actor))
            self.assertEquals([(0,0),(4,4),(0,2),(2,0)], actor.path)
            self.assertEquals('1.84,0.00,-0.00', self._fmt(tuple(actor.actor_obj.pos)))
            self.assertEquals([action1, action2Redo, action3], actor.actions)

        def eb(reason):
            self.fail('Error : %s' % reason)

        action2Redo = HappyAction(actor, 0, 2)
        d = actor.act(action2Redo, index=1).addCallbacks(cb, eb)
        action2Redo.update(0.0123571287)
        action3.update(0.031246741627)
        return d


    def testPositioning(self):

        actor = roboto.actor.Actor(mesh.loadObj(P('data', 'lilbot.obj')))
        actor.start((0,0))
        self.assertEquals(0, actor.row)
        self.assertEquals(0, actor.col)
        self.assertEquals('0.00,0.00,-0.00', self._fmt(tuple(actor.actor_obj.pos)))

        action1 = HappyAction(actor, 4, 4)
        actor.act(action1)
        action1.update(0.04327647)
        self.assertEquals(4, actor.row)
        self.assertEquals(4, actor.col)
        self.assertEquals('3.67,0.00,-3.67', self._fmt(tuple(actor.actor_obj.pos)))

        action2 = HappyAction(actor, 1, 3)
        actor.act(action2)
        action2.update(0.03426756)
        self.assertEquals(1, actor.row)
        self.assertEquals(3, actor.col)
        self.assertEquals('0.92,0.00,-2.75', self._fmt(tuple(actor.actor_obj.pos)))
        

    def _fmt(self, t):
        return ','.join([('%2.2f' % n) for n in t])

    def testRollback(self):
        actor = roboto.actor.Actor(mesh.loadObj(P('data', 'lilbot.obj')))
        actor.start((0,0))
        
        action1 = HappyAction(actor, 4, 4)
        actor.act(action1)
        action1.update(0.04327647)
        
        action2 = HappyAction(actor, 1, 3)
        actor.act(action2)
        action2.update(0.03426756)
        

        def cb(the_actor):
            self.assertEquals(id(the_actor), id(actor))
            self.assertEquals([(0,0),(4,4)], actor.path)
            self.assertEquals([action1], actor.actions)
            self.assertEquals(4, actor.row)
            self.assertEquals(4, actor.col)
            self.assertEquals('3.67,0.00,-3.67', self._fmt(tuple(actor.actor_obj.pos)))

        def eb(reason):
            self.fail('Error : %s' % reason)

        actor.rollback().addCallbacks(cb, eb)


    def testDeleteFromPath(self):
        actor = roboto.actor.Actor(mesh.loadObj(P('data', 'lilbot.obj')))
        actor.start((0,0))
        
        action1 = HappyAction(actor, 1,1)
        action2 = HappyAction(actor, 2,2)
        action3 = HappyAction(actor, 3,3)
        
        actor.act(action1)
        action1.update(0.032647127)
        actor.act(action2)
        action2.update(0.021347612)
        actor.act(action3)
        action3.update(0.014736137)

        actor.act(None, index=1)
        action3.update(0.04327167)
        self.assertEquals([(0,0),(1,1),(3,3)], actor.path)
        self.assertEquals([action1, action3], actor.actions)
        log.msg("self.assertEquals('%2.2f,%2.2f,%2.2f', self._fmt(tuple(actor.actor_obj.pos)))" % tuple(actor.actor_obj.pos))

        

    def testArrgggggg(self):
        actor = roboto.actor.Actor(mesh.loadObj(P('data', 'lilbot.obj')))
        actor.start((0,0))
        
        # The user types in 5 commands
        
        action1 = HappyAction(actor, 1,1)
        action2 = HappyAction(actor, 2,2)
        action3 = HappyAction(actor, 3,3)
        action4 = HappyAction(actor, 4,4)
        action5 = HappyAction(actor, 0,2)

        actor.act(action1)
        action1.update(0.032647127)
        actor.act(action2)
        action2.update(0.021347612)
        actor.act(action3)
        action3.update(0.014736137)
        actor.act(action4)
        action4.update(0.014736137)
        actor.act(action5)
        action5.update(0.021347382)
        
        self.assertEquals([(0,0),(1,1),(2,2),(3,3),(4,4),(0,2)], actor.path)

        # We emulate this weird scenario
        # 1. The user goes back to the action 3 and edits it
        # 2. Then they skip forward and edit the action 4
        # 3. ... they skip back and edit action 2
        # 4. Then they go back to action 4 and delete it
        # 5. Then tehy go to action 5 and edit it

        action3Edit = HappyAction(actor, 2, 3)
        actor.act(action3Edit, index=2)
        action4Edit = HappyAction(actor, 4, 0)
        actor.act(action4Edit, index=3)
        action2Edit = HappyAction(actor, 0, 3)
        actor.act(action2Edit, index=1)
        actor.act(None, index=3)
        action5Edit = HappyAction(actor, 2,1)
        actor.act(action5Edit, index=3) # index became 3

        # ... now just action3Edit, action2Edit
        # Get run ...
        action3Edit.update(0.013424547)
        #action2Edit.update(0.012432154)

        #self.assertEquals([(0,0),(1,1),(0,3),(2,3),(2,1)], actor.path)
        self.assertEquals([action1,action2Edit,action3Edit,action5Edit], actor.actions)


